# Copyright © 2023 Intel Corporation
# SPDX-License-Identifier: MIT

import sys

from builtin_types import BUILTIN_TYPES
from mako.template import Template

template = """\
/*
 * Copyright 2023 Intel Corporation
 * SPDX-License-Identifier: MIT
 */

/* This is an automatically generated file. */

#ifndef _BUILTIN_TYPES_
#define _BUILTIN_TYPES_

%for t in BUILTIN_TYPES:
extern const struct glsl_type glsl_type_builtin_${t["name"]};
%endfor

#endif /* _BUILTIN_TYPES_ */
"""

if len(sys.argv) < 2:
    print('Missing output argument', file=sys.stderr)
    sys.exit(1)

output = sys.argv[1]

with open(output, 'w', encoding='utf-8') as f:
    f.write(Template(template).render(BUILTIN_TYPES=BUILTIN_TYPES))



#encoding=utf-8

# Copyright (C) 2021 Collabora, Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from valhall import valhall_parse_isa

(_, _, enums, _, safe_name) = valhall_parse_isa()

print("#ifndef __VALHALL_ENUMS_H_")
print("#define __VALHALL_ENUMS_H_")

for enum in sorted(enums):
    print(f"enum va_{safe_name(enum)} {{")

    for i, value in enumerate(enums[enum].values):
        if value.value != 'reserved':
            key = safe_name(f"va_{enum}_{value.value}")
            print(f"   {key.upper()} = {i},")

    print("};\n")

print("#endif")



from .common.codegen import CodeGen, VulkanWrapperGenerator
from .common.vulkantypes import \
        VulkanAPI, makeVulkanTypeSimple, iterateVulkanType
from .common.vulkantypes import EXCLUDED_APIS
from .common.vulkantypes import HANDLE_TYPES

import copy
import re

RESOURCE_TRACKER_ENTRIES = [
    "vkEnumerateInstanceExtensionProperties",
    "vkEnumerateDeviceExtensionProperties",
    "vkEnumeratePhysicalDevices",
    "vkAllocateMemory",
    "vkFreeMemory",
    "vkCreateImage",
    "vkDestroyImage",
    "vkGetImageMemoryRequirements",
    "vkGetImageMemoryRequirements2",
    "vkGetImageMemoryRequirements2KHR",
    "vkGetImageDrmFormatModifierPropertiesEXT",
    "vkBindImageMemory",
    "vkBindImageMemory2",
    "vkBindImageMemory2KHR",
    "vkCreateBuffer",
    "vkDestroyBuffer",
    "vkGetBufferMemoryRequirements",
    "vkGetBufferMemoryRequirements2",
    "vkGetBufferMemoryRequirements2KHR",
    "vkBindBufferMemory",
    "vkBindBufferMemory2",
    "vkBindBufferMemory2KHR",
    "vkCreateSemaphore",
    "vkDestroySemaphore",
    "vkQueueSubmit",
    "vkQueueSubmit2",
    "vkQueueWaitIdle",
    "vkImportSemaphoreFdKHR",
    "vkGetSemaphoreFdKHR",
    # Warning: These need to be defined in vk.xml (currently no-op) {
    "vkGetMemoryFuchsiaHandleKHR",
    "vkGetMemoryFuchsiaHandlePropertiesKHR",
    "vkGetSemaphoreFuchsiaHandleKHR",
    "vkImportSemaphoreFuchsiaHandleKHR",
    # } end Warning: These need to be defined in vk.xml (currently no-op)
    "vkGetAndroidHardwareBufferPropertiesANDROID",
    "vkGetMemoryAndroidHardwareBufferANDROID",
    "vkGetMemoryFdKHR",
    "vkGetMemoryFdPropertiesKHR",
    "vkCreateSamplerYcbcrConversion",
    "vkDestroySamplerYcbcrConversion",
    "vkCreateSamplerYcbcrConversionKHR",
    "vkDestroySamplerYcbcrConversionKHR",
    "vkUpdateDescriptorSetWithTemplate",
    "vkUpdateDescriptorSetWithTemplateKHR",
    "vkGetPhysicalDeviceFormatProperties2",
    "vkGetPhysicalDeviceFormatProperties2KHR",
    "vkGetPhysicalDeviceImageFormatProperties2",
    "vkGetPhysicalDeviceImageFormatProperties2KHR",
    "vkBeginCommandBuffer",
    "vkEndCommandBuffer",
    "vkResetCommandBuffer",
    "vkCreateImageView",
    "vkCreateSampler",
    "vkGetPhysicalDeviceExternalFenceProperties",
    "vkGetPhysicalDeviceExternalFencePropertiesKHR",
    "vkGetPhysicalDeviceExternalBufferProperties",
    "vkGetPhysicalDeviceExternalBufferPropertiesKHR",
    "vkCreateFence",
    "vkResetFences",
    "vkImportFenceFdKHR",
    "vkGetFenceFdKHR",
    "vkGetFenceStatus",
    "vkWaitForFences",
    "vkCreateDescriptorPool",
    "vkDestroyDescriptorPool",
    "vkResetDescriptorPool",
    "vkAllocateDescriptorSets",
    "vkFreeDescriptorSets",
    "vkCreateDescriptorSetLayout",
    "vkCmdExecuteCommands",
    "vkCmdBindDescriptorSets",
    "vkDestroyDescriptorSetLayout",
    "vkAllocateCommandBuffers",
    "vkQueueSignalReleaseImageANDROID",
    "vkCmdPipelineBarrier",
    "vkCreateGraphicsPipelines",
    "vkCmdClearColorImage",
    # Fuchsia
    "vkGetMemoryZirconHandleFUCHSIA",
    "vkGetMemoryZirconHandlePropertiesFUCHSIA",
    "vkGetSemaphoreZirconHandleFUCHSIA",
    "vkImportSemaphoreZirconHandleFUCHSIA",
    "vkCreateBufferCollectionFUCHSIA",
    "vkDestroyBufferCollectionFUCHSIA",
    "vkSetBufferCollectionImageConstraintsFUCHSIA",
    "vkSetBufferCollectionBufferConstraintsFUCHSIA",
    "vkGetBufferCollectionPropertiesFUCHSIA",
    "vkSetPrivateData",
    "vkSetPrivateDataKHR",
    "vkGetPrivateData",
    "vkGetPrivateDataKHR",
    "vkCreatePrivateDataSlot",
    "vkCreatePrivateDataSlotEXT",
    "vkDestroyPrivateDataSlot",
    "vkDestroyPrivateDataSlotEXT",
]

SUCCESS_VAL = {
    "VkResult" : ["VK_SUCCESS"],
}

HANDWRITTEN_ENTRY_POINTS = [
    # Instance/device/physical-device special-handling, dispatch tables, etc..
    "vkCreateInstance",
    "vkDestroyInstance",
    "vkGetInstanceProcAddr",
    "vkEnumerateInstanceVersion",
    "vkEnumerateInstanceLayerProperties",
    "vkEnumerateInstanceExtensionProperties",
    "vkEnumerateDeviceExtensionProperties",
    "vkGetDeviceProcAddr",
    "vkEnumeratePhysicalDevices",
    "vkEnumeratePhysicalDeviceGroups",
    "vkCreateDevice",
    "vkDestroyDevice",
    # Manual alloc/free + vk_*_init/free() call w/ special params
    "vkGetDeviceQueue",
    "vkGetDeviceQueue2",
    # Command pool/buffer handling
    "vkCreateCommandPool",
    "vkDestroyCommandPool",
    "vkAllocateCommandBuffers",
    "vkResetCommandPool",
    "vkFreeCommandBuffers",
    "vkResetCommandPool",
    # Transform feedback
    "vkCmdBeginTransformFeedbackEXT",
    # Special cases to handle struct translations in the pNext chain
    # TODO: Make a codegen module (use deepcopy as reference) to make this more robust
    "vkAllocateMemory",
    "vkUpdateDescriptorSets",
]

# Handles that need to be translated to/from their corresponding gfxstream object types
HANDLES_TRANSLATE = {
    "VkInstance",
    "VkPhysicalDevice",
    "VkDevice",
    "VkQueue",
    "VkCommandPool",
    "VkCommandBuffer",
    "VkFence",
    "VkSemaphore",
    # TODO: Still need this translation to avoid descriptorSets crash
    "VkBuffer",
}

# Types that have a corresponding method for transforming
# an input list to its internal counterpart
TYPES_TRANSFORM_LIST_METHOD = {
    "VkFence",
    "VkSemaphore",
    "VkSemaphoreSubmitInfo",
}

def is_cmdbuf_dispatch(api):
    return "VkCommandBuffer" == api.parameters[0].typeName

def is_queue_dispatch(api):
    return "VkQueue" == api.parameters[0].typeName

def getCreateParam(api):
    for param in api.parameters:
        if param.isCreatedBy(api):
            return param
    return None

def getDestroyParam(api):
    for param in api.parameters:
        if param.isDestroyedBy(api):
            return param
    return None

# i.e. VkQueryPool --> vk_query_pool
def typeNameToMesaType(typeName):
    vkTypeNameRegex = "(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])"
    words = re.split(vkTypeNameRegex, typeName)
    outputType = "vk"
    for word in words[1:]:
        outputType += "_"
        outputType += word.lower()
    return outputType

def typeNameToBaseName(typeName):
    return typeNameToMesaType(typeName)[len("vk_"):]

def paramNameToObjectName(paramName):
    return "gfxstream_%s" % paramName

def typeNameToVkObjectType(typeName):
    return "VK_OBJECT_TYPE_%s" % typeNameToBaseName(typeName).upper()

def typeNameToObjectType(typeName):
    return "gfxstream_vk_%s" % typeNameToBaseName(typeName)

def transformListFuncName(typeName):
    return "transform%sList" % (typeName)

def isAllocatorParam(param):
    ALLOCATOR_TYPE_NAME = "VkAllocationCallbacks"
    return (param.pointerIndirectionLevels == 1
            and param.isConst
            and param.typeName == ALLOCATOR_TYPE_NAME)

def isArrayParam(param):
    return (1 == param.pointerIndirectionLevels
            and param.isConst
            and "len" in param.attribs)

INTERNAL_OBJECT_NAME = "internal_object"

class VulkanFuncTable(VulkanWrapperGenerator):
    def __init__(self, module, typeInfo):
        VulkanWrapperGenerator.__init__(self, module, typeInfo)
        self.typeInfo = typeInfo
        self.cgen = CodeGen()
        self.entries = []
        self.entryFeatures = []
        self.cmdToFeatureType = {}
        self.feature = None
        self.featureType = None

    def onBegin(self,):
        cgen = self.cgen
        self.module.appendImpl(cgen.swapCode())
        pass

    def onBeginFeature(self, featureName, featureType):
        self.feature = featureName
        self.featureType = featureType

    def onEndFeature(self):
        self.feature = None
        self.featureType = None

    def onFeatureNewCmd(self, name):
        self.cmdToFeatureType[name] = self.featureType

    def onGenCmd(self, cmdinfo, name, alias):
        typeInfo = self.typeInfo
        cgen = self.cgen
        api = typeInfo.apis[name]
        self.entries.append(api)
        self.entryFeatures.append(self.feature)
        self.loopVars = ["i", "j", "k", "l", "m", "n"]
        self.loopVarIndex = 0

        def getNextLoopVar():
            if self.loopVarIndex >= len(self.loopVars):
                raise
            loopVar = self.loopVars[self.loopVarIndex]
            self.loopVarIndex += 1
            return loopVar

        def isCompoundType(typeName):
            return typeInfo.isCompoundType(typeName)

        def handleTranslationRequired(typeName):
            return typeName in HANDLE_TYPES and typeName in HANDLES_TRANSLATE

        def translationRequired(typeName):
            if isCompoundType(typeName):
                struct = typeInfo.structs[typeName]
                for member in struct.members:
                    if translationRequired(member.typeName):
                        return True
                return False
            else:
                return handleTranslationRequired(typeName)

        def genDestroyGfxstreamObjects():
            destroyParam = getDestroyParam(api)
            if not destroyParam:
                return
            if not translationRequired(destroyParam.typeName):
                return
            objectName = paramNameToObjectName(destroyParam.paramName)
            allocatorParam = "NULL"
            for p in api.parameters:
                if isAllocatorParam(p):
                    allocatorParam = p.paramName
            deviceParam = api.parameters[0]
            if "VkDevice" != deviceParam.typeName:
                print("ERROR: Unhandled non-VkDevice parameters[0]: %s (for API: %s)" %(deviceParam.typeName, api.name))
                raise
            # call vk_object_free() directly
            mesaObjectDestroy = "(void *)%s" % objectName
            cgen.funcCall(
                None,
                "vk_object_free",
                ["&%s->vk" % paramNameToObjectName(deviceParam.paramName), allocatorParam, mesaObjectDestroy]
            )

        def genMesaObjectAlloc(allocCallLhs):
            deviceParam = api.parameters[0]
            if "VkDevice" != deviceParam.typeName:
                print("ERROR: Unhandled non-VkDevice parameters[0]: %s (for API: %s)" %(deviceParam.typeName, api.name))
                raise
            allocatorParam = "NULL"
            for p in api.parameters:
                if isAllocatorParam(p):
                    allocatorParam = p.paramName
            createParam = getCreateParam(api)
            objectType = typeNameToObjectType(createParam.typeName)
            # Call vk_object_zalloc directly
            cgen.funcCall(
                allocCallLhs,
                "(%s *)vk_object_zalloc" % objectType,
                ["&%s->vk" % paramNameToObjectName(deviceParam.paramName), allocatorParam, ("sizeof(%s)" % objectType), typeNameToVkObjectType(createParam.typeName)]
            )

        # Alloc/create gfxstream_vk_* object
        def genCreateGfxstreamObjects():
            createParam = getCreateParam(api)
            if not createParam:
                return False
            if not handleTranslationRequired(createParam.typeName):
                return False
            objectType = "struct %s" % typeNameToObjectType(createParam.typeName)
            callLhs = "%s *%s" % (objectType, paramNameToObjectName(createParam.paramName))
            genMesaObjectAlloc(callLhs)

            retVar = api.getRetVarExpr()
            if retVar:
                retTypeName = api.getRetTypeExpr()
                # ex: vkCreateBuffer_VkResult_return = gfxstream_buffer ? VK_SUCCESS : VK_ERROR_OUT_OF_HOST_MEMORY;
                cgen.stmt("%s = %s ? %s : %s" %
                          (retVar, paramNameToObjectName(createParam.paramName), SUCCESS_VAL[retTypeName][0], "VK_ERROR_OUT_OF_HOST_MEMORY"))
            return True

        def genVkFromHandle(param, fromName):
            objectName = paramNameToObjectName(param.paramName)
            cgen.stmt("VK_FROM_HANDLE(%s, %s, %s)" %
                      (typeNameToObjectType(param.typeName), objectName, fromName))
            return objectName

        def genGetGfxstreamHandles():
            createParam = getCreateParam(api)
            for param in api.parameters:
                if not handleTranslationRequired(param.typeName):
                    continue
                elif isArrayParam(param):
                    continue
                elif param != createParam:
                    if param.pointerIndirectionLevels > 0:
                        print("ERROR: Unhandled pointerIndirectionLevels > 1 for API %s (param %s)" % (api.name, param.paramName))
                        raise
                    genVkFromHandle(param, param.paramName)

        def internalNestedParamName(param):
            parentName = ""
            if param.parent:
                parentName = "_%s" % param.parent.typeName
            return "internal%s_%s" % (parentName, param.paramName)

        def genInternalArrayDeclarations(param, countParamName, nestLevel=0):
            internalArray = None
            if 0 == nestLevel:
                internalArray = "internal_%s" % param.paramName
                cgen.stmt("std::vector<%s> %s(%s)" % (param.typeName, internalArray, countParamName))
            elif 1 == nestLevel or 2 == nestLevel:
                internalArray = internalNestedParamName(param)
                if isArrayParam(param):
                    cgen.stmt("std::vector<std::vector<%s>> %s" % (param.typeName, internalArray))
                else:
                    cgen.stmt("std::vector<%s> %s" % (param.typeName, internalArray))
            else:
                print("ERROR: nestLevel > 2 not verified.")
                raise
            if isCompoundType(param.typeName):
                for member in typeInfo.structs[param.typeName].members:
                    if translationRequired(member.typeName):
                        if handleTranslationRequired(member.typeName) and not isArrayParam(member):
                            # No declarations for non-array handleType
                            continue
                        genInternalArrayDeclarations(member, countParamName, nestLevel + 1)
            return internalArray

        def genInternalCompoundType(param, outName, inName, currLoopVar):
            nextLoopVar = None
            cgen.stmt("%s = %s" % (outName, inName))
            for member in typeInfo.structs[param.typeName].members:
                if not translationRequired(member.typeName):
                    continue
                cgen.line("/* %s::%s */" % (param.typeName, member.paramName))
                nestedOutName = ("%s[%s]" % (internalNestedParamName(member), currLoopVar))
                if isArrayParam(member):
                    countParamName = "%s.%s" % (outName, member.attribs["len"])
                    inArrayName = "%s.%s" % (outName, member.paramName)
                    cgen.stmt("%s.push_back(std::vector<%s>())" % (internalNestedParamName(member), member.typeName))
                    if member.typeName in TYPES_TRANSFORM_LIST_METHOD:
                        # Use the corresponding transformList call
                        cgen.funcCall(nestedOutName, transformListFuncName(member.typeName), [inArrayName, countParamName])
                        cgen.stmt("%s = %s.data()" % (inArrayName, nestedOutName))
                        cgen.stmt("%s = %s.size()" % (countParamName, nestedOutName))
                    else:
                        # Standard translation
                        cgen.stmt("%s.resize(%s)" % (nestedOutName, countParamName))
                        if not nextLoopVar:
                            nextLoopVar = getNextLoopVar()
                        internalArray = genInternalArray(member, countParamName, nestedOutName, inArrayName, nextLoopVar)
                        cgen.stmt("%s = %s" %(inArrayName, internalArray))
                elif isCompoundType(member.typeName):
                    memberFullName = "%s.%s" % (outName, member.paramName)
                    if 1 == member.pointerIndirectionLevels:
                        cgen.beginIf(memberFullName)
                        inParamName = "%s[0]" % memberFullName
                        genInternalCompoundType(member, nestedOutName, inParamName, currLoopVar)
                        cgen.stmt("%s.%s = &%s" % (outName, member.paramName,  nestedOutName))
                    else:
                        cgen.beginBlock()
                        genInternalCompoundType(member, nestedOutName, memberFullName, currLoopVar)
                        cgen.stmt("%s.%s = %s" % (outName, member.paramName,  nestedOutName))
                    cgen.endBlock()
                else:
                    # Replace member with internal object
                    replaceName = "%s.%s" % (outName, member.paramName)
                    if member.isOptional:
                        cgen.beginIf(replaceName)
                    gfxstreamObject = genVkFromHandle(member, replaceName)
                    cgen.stmt("%s = %s->%s" % (replaceName, gfxstreamObject, INTERNAL_OBJECT_NAME))
                    if member.isOptional:
                        cgen.endIf()

        def genInternalArray(param, countParamName, outArrayName, inArrayName, loopVar):
            if param.typeName in TYPES_TRANSFORM_LIST_METHOD:
                # Use the corresponding transformList call
                cgen.funcCall(outArrayName, transformListFuncName(param.typeName), [inArrayName, countParamName])
                cgen.stmt("%s = %s.data()" % (inArrayName, outArrayName))
                cgen.stmt("%s = %s.size()" % (countParamName, outArrayName))
            else:
                cgen.beginFor("uint32_t %s = 0" % loopVar, "%s < %s" % (loopVar, countParamName), "++%s" % loopVar)
                if param.isOptional:
                    cgen.beginIf(inArrayName)
                if isCompoundType(param.typeName):
                    genInternalCompoundType(param, ("%s[%s]" % (outArrayName, loopVar)), "%s[%s]" % (inArrayName, loopVar), loopVar)
                else:
                    gfxstreamObject = genVkFromHandle(param, "%s[%s]" % (inArrayName, loopVar))
                    cgen.stmt("%s[%s] = %s->%s" % (outArrayName, loopVar, gfxstreamObject, INTERNAL_OBJECT_NAME))
                if param.isOptional:
                    cgen.endIf()
                cgen.endFor()
            return "%s.data()" % outArrayName

        # Translate params into params needed for gfxstream-internal
        #  encoder/resource-tracker calls
        def getEncoderOrResourceTrackerParams():
            createParam = getCreateParam(api)
            outParams = copy.deepcopy(api.parameters)
            nextLoopVar = getNextLoopVar()
            for param in outParams:
                if not translationRequired(param.typeName):
                    continue
                elif isArrayParam(param) or isCompoundType(param.typeName):
                    if param.possiblyOutput():
                        print("ERROR: Unhandled CompoundType / Array output for API %s (param %s)" % (api.name, param.paramName))
                        raise
                    if 1 != param.pointerIndirectionLevels or not param.isConst:
                        print("ERROR: Compound type / array input is not 'const <type>*' (API: %s, paramName: %s)" % (api.name, param.paramName))
                        raise
                    countParamName = "1"
                    if "len" in param.attribs:
                        countParamName = param.attribs["len"]
                    internalArrayName = genInternalArrayDeclarations(param, countParamName)
                    param.paramName = genInternalArray(param, countParamName, internalArrayName, param.paramName, nextLoopVar)
                elif 0 == param.pointerIndirectionLevels:
                    if param.isOptional:
                        param.paramName = ("%s ? %s->%s : VK_NULL_HANDLE" % (paramNameToObjectName(param.paramName), paramNameToObjectName(param.paramName), INTERNAL_OBJECT_NAME))
                    else:
                        param.paramName = ("%s->%s" % (paramNameToObjectName(param.paramName), INTERNAL_OBJECT_NAME))
                elif createParam and param.paramName == createParam.paramName:
                    param.paramName = ("&%s->%s" % (paramNameToObjectName(param.paramName), INTERNAL_OBJECT_NAME))
                else:
                    print("ERROR: Unknown handling for param: %s (API: %s)" % (param, api.name))
                    raise
            return outParams

        def genEncoderOrResourceTrackerCall(declareResources=True):
            if is_cmdbuf_dispatch(api):
                cgen.stmt("auto vkEnc = gfxstream::vk::ResourceTracker::getCommandBufferEncoder(%s->%s)" % (paramNameToObjectName(api.parameters[0].paramName), INTERNAL_OBJECT_NAME))
            elif is_queue_dispatch(api):
                cgen.stmt("auto vkEnc = gfxstream::vk::ResourceTracker::getQueueEncoder(%s->%s)" % (paramNameToObjectName(api.parameters[0].paramName), INTERNAL_OBJECT_NAME))
            else:
                cgen.stmt("auto vkEnc = gfxstream::vk::ResourceTracker::getThreadLocalEncoder()")
            callLhs = None
            retTypeName = api.getRetTypeExpr()
            if retTypeName != "void":
                callLhs = api.getRetVarExpr()

            # Get parameter list modded for gfxstream-internal call
            parameters = getEncoderOrResourceTrackerParams()
            if name in RESOURCE_TRACKER_ENTRIES:
                if declareResources:
                    cgen.stmt("auto resources = gfxstream::vk::ResourceTracker::get()")
                cgen.funcCall(
                    callLhs, "resources->" + "on_" + api.name,
                    ["vkEnc"] + SUCCESS_VAL.get(retTypeName, []) + \
                    [p.paramName for p in parameters])
            else:
                cgen.funcCall(
                    callLhs, "vkEnc->" + api.name, [p.paramName for p in parameters] + ["true /* do lock */"])

        def genReturnExpression():
            retTypeName = api.getRetTypeExpr()
            # Set the createParam output, if applicable
            createParam = getCreateParam(api)
            if createParam and handleTranslationRequired(createParam.typeName):
                if 1 != createParam.pointerIndirectionLevels:
                    print("ERROR: Unhandled pointerIndirectionLevels != 1 in return for API %s (createParam %s)" % api.name, createParam.paramName)
                    raise
                # ex: *pBuffer = gfxstream_vk_buffer_to_handle(gfxstream_buffer)
                cgen.funcCall(
                    "*%s" % createParam.paramName,
                    "%s_to_handle" % typeNameToObjectType(createParam.typeName),
                    [paramNameToObjectName(createParam.paramName)]
                )

            if retTypeName != "void":
                cgen.stmt("return %s" % api.getRetVarExpr())

        def genGfxstreamEntry(declareResources=True):
            cgen.stmt("MESA_TRACE_SCOPE(\"%s\")" % api.name)
            # declare returnVar
            retTypeName = api.getRetTypeExpr()
            retVar = api.getRetVarExpr()
            if retVar:
                cgen.stmt("%s %s = (%s)0" % (retTypeName, retVar, retTypeName))
            # Check non-null destroy param for free/destroy calls
            destroyParam = getDestroyParam(api)
            if destroyParam:
                cgen.beginIf("VK_NULL_HANDLE == %s" % destroyParam.paramName)
                if api.getRetTypeExpr() != "void":
                    cgen.stmt("return %s" % api.getRetVarExpr())
                else:
                    cgen.stmt("return")
                cgen.endIf()
            # Translate handles
            genGetGfxstreamHandles()
            # Translation/creation of objects
            createdObject = genCreateGfxstreamObjects()
            # Make encoder/resource-tracker call
            if retVar and createdObject:
                cgen.beginIf("%s == %s" % (SUCCESS_VAL[retTypeName][0], retVar))
            else:
                cgen.beginBlock()
            genEncoderOrResourceTrackerCall()
            cgen.endBlock()
            # Destroy gfxstream objects
            genDestroyGfxstreamObjects()
            # Set output / return variables
            genReturnExpression()

        api_entry = api.withModifiedName("gfxstream_vk_" + api.name[2:])
        if api.name not in HANDWRITTEN_ENTRY_POINTS:
            cgen.line(self.cgen.makeFuncProto(api_entry))
            cgen.beginBlock()
            genGfxstreamEntry()
            cgen.endBlock()
            self.module.appendImpl(cgen.swapCode())


    def onEnd(self,):
        pass

    def isDeviceDispatch(self, api):
        # TODO(230793667): improve the heuristic and just use "cmdToFeatureType"
        return (len(api.parameters) > 0 and
            "VkDevice" == api.parameters[0].typeName) or (
            "VkCommandBuffer" == api.parameters[0].typeName and
            self.cmdToFeatureType.get(api.name, "") == "device")



# Copyright © 2025 Collabora Ltd.
# SPDX-License-Identifier: MIT

TEMPLATE = """
#include "bi_swizzles.h"

uint32_t bi_op_swizzles[BI_NUM_OPCODES][BI_MAX_SRCS] = {
% for opcode, src_swizzles in op_swizzles.items():
    [BI_OPCODE_${opcode.replace('.', '_').upper()}] = {
% for src in src_swizzles:
% for swizzle in src:
        (1 << BI_SWIZZLE_${swizzle.upper()}) |
% endfor
        0,
% endfor
    },
% endfor
};
"""

import sys
from bifrost_isa import *
from mako.template import Template

instructions = {}
for arg in sys.argv[1:]:
    new_instructions = parse_instructions(arg, include_pseudo = True)
    instructions.update(new_instructions)

ir_instructions = partition_mnemonics(instructions)

op_swizzles = {}

for name, op in ir_instructions.items():
    src_swizzles = [None] * op['srcs']
    SWIZZLE_MODS = ["lane", "lanes", "replicate", "swz", "widen", "swap"]

    for mod, opts in op['modifiers'].items():
        raw, arg = (mod[0:-1], int(mod[-1])) if mod[-1] in "0123" else (mod, 0)
        if raw in SWIZZLE_MODS:
            assert(src_swizzles[arg] is None)
            src_swizzles[arg] = set(opts)

            src_swizzles[arg].discard('reserved')
            if 'none' in src_swizzles[arg]:
                src_swizzles[arg].remove('none')
                src_swizzles[arg].add('h01')

    # Anything that doesn't have swizzle mods only supports identity
    src_swizzles = [swizzle or ['h01'] for swizzle in src_swizzles]

    op_swizzles[name] = src_swizzles

print(Template(COPYRIGHT + TEMPLATE).render(op_swizzles = op_swizzles))



#!/usr/bin/env python3
#
# Copyright © 2024 Igalia S.L.
# SPDX-License-Identifier: MIT

import argparse
import ctypes
import logging
import re
from datetime import datetime
from mako.template import Template
from pycparser import parse_file, c_ast

logger = logging.getLogger('hwdb')

template = """/*
 * Copyright © 2024 Igalia S.L.
 * SPDX-License-Identifier: MIT
 */

#pragma once

#include <stdint.h>

typedef struct
{
% for type, name in struct:
   ${type} ${name};
% endfor
} gcsFEATURE_DATABASE;

static gcsFEATURE_DATABASE gChipInfo[] = {
% for entry in entries:
   {
% for name, value in entry:
      ${value}, /* ${name} */
% endfor
   },
% endfor
};

static inline gcsFEATURE_DATABASE *
gcQueryFeatureDB(
    uint32_t ChipID,
    uint32_t ChipVersion,
    uint32_t ProductID,
    uint32_t EcoID,
    uint32_t CustomerID
    )
{
    int entryNum = sizeof(gChipInfo) / sizeof(gChipInfo[0]);

    /* check formal release entries first */
    for (int i = 0; i < entryNum; ++i)
    {
        if ((gChipInfo[i].chipID == ChipID)
            && (gChipInfo[i].chipVersion == ChipVersion)
            && (gChipInfo[i].productID == ProductID)
            && (gChipInfo[i].ecoID == EcoID)
            && (gChipInfo[i].customerID == CustomerID)
            && (gChipInfo[i].formalRelease)
           )
        {
            return &gChipInfo[i];
        }
    }

    /* check informal release entries if we dont find in formal entries */
    for (int i = 0; i < entryNum; ++i)
    {
        if ((gChipInfo[i].chipID == ChipID)
            && ((gChipInfo[i].chipVersion & 0xFFF0) == (ChipVersion & 0xFFF0))
            && (gChipInfo[i].productID == ProductID)
            && (gChipInfo[i].ecoID == EcoID)
            && (gChipInfo[i].customerID == CustomerID)
            && (!gChipInfo[i].formalRelease)
           )
        {
            return &gChipInfo[i];
        }
    }

    return 0;
}
"""


class HeaderFile(c_ast.NodeVisitor):
    """Class representing a complete header file"""

    # Regular expression to match the date and time in the comment
    _DATE_RE = re.compile(r'/\*Auto created on (\d{4}-\d{2}-\d{2} \d{2}:\d{2})\*/')

    def __init__(self, filename):
        self.filename = filename
        self.structs = {}
        self.data = []
        self.date_time = None
        self.database_struct = None

        self._read_date()
        self._parse()

        logger.debug('Parsed %s (autogenerated at %s, %u struct members, %u entries)', self.filename, self.date_time, len(self.database_struct._fields_), len(self.data))

    def _read_date(self):
        """Function parsing the creation date with re."""
        # Read the content of the file and search for pattern
        with open(self.filename, 'r', encoding="utf-8") as file:
            file_content = file.read()

        match = self._DATE_RE.search(file_content)

        if match:
            self.date_time = datetime.strptime(match.group(1), '%Y-%m-%d %H:%M')

    def _parse(self):
        ast = parse_file(self.filename, use_cpp=True, cpp_args=['-E', r'-I./utils/fake_libc_include', '-DgctUINT32=unsigned int', '-DgctINT=int'])
        self.visit(ast)

        self.database_struct = self.structs['gcsFEATURE_DATABASE']

    def visit_Typedef(self, node):
        if isinstance(node.type, c_ast.TypeDecl) and isinstance(node.type.type, c_ast.Struct):
            struct_node = node.type.type
            struct_name = node.name  # Typedef name as the struct name
            fields = self._extract_fields_from_struct(struct_node)
            if fields is not None:
                # Create the ctypes.Structure subclass and add it to the structures dictionary
                self.structs[struct_name] = type(struct_name, (ctypes.Structure,), {'_fields_': fields})

    def _extract_fields_from_struct(self, struct_node):
        """Function returning all fields of struct."""
        fields = []
        for decl in (struct_node.decls or []):
            if isinstance(decl.type, c_ast.TypeDecl) or isinstance(decl.type, c_ast.PtrDecl):
                field_name = decl.name
                field_type = self._map_type_to_ctypes(decl.type.type, decl.bitsize)
                if field_type:
                    fields.append((field_name, field_type))
            elif isinstance(decl.type, c_ast.ArrayDecl):
                # Handle array type
                field_name = decl.type.type.declname
                element_type = self._map_type_to_ctypes(decl.type.type.type)
                array_size = int(decl.type.dim.value)  # Assuming dim is a Constant node with the size as its value
                if element_type:
                    fields.append((field_name, element_type * array_size))

        return fields if fields else None

    def _map_type_to_ctypes(self, type_node, bitsize=None):
        """Function returning a ctype type based node type."""
        type_mappings = {
            'bool': ctypes.c_bool,
            'unsigned int': ctypes.c_uint,
            'int': ctypes.c_int,
        }

        if isinstance(type_node, c_ast.IdentifierType):
            c_type = ' '.join(type_node.names)

            if bitsize and bitsize.value == '1':
                c_type = 'bool'

            return type_mappings.get(c_type)
        elif isinstance(type_node, c_ast.TypeDecl):
            return ctypes.c_char_p

        return None

    def visit_Decl(self, node):
        # Check if the node is a declaration of an array of structs
        if isinstance(node.type, c_ast.ArrayDecl) and isinstance(node.type.type, c_ast.TypeDecl):
            struct_name = node.type.type.type.names[0]
            if struct_name in self.structs:
                elements = self._parse_array_initializer(node.init, self.structs[struct_name])
                self.data.extend(elements)

    def _parse_initializer(self, initializer, struct):
        """Function returning one parsed struct initializer."""
        return [
            (param if not isinstance(param, str) else param.encode('utf-8'))
            for index, expr in enumerate(initializer.exprs)
            for param in [self._parse_expr(expr, getattr(struct._fields_[index][1], '_length_', None))]
        ]

    def _parse_array_initializer(self, init, struct_class):
        """Function returning a fully processed struct initializer list."""
        assert (isinstance(init, c_ast.InitList))
        return [struct_class(*self._parse_initializer(initializer, struct_class)) for initializer in init.exprs]

    def _parse_expr(self, expr, expected_size=None):
        """Function returning parsed expression."""
        # Direct handling of constant types
        if isinstance(expr, c_ast.Constant):
            if expr.type == "int":
                # Base 0 automatically handles decimal, hex, and octal
                return int(expr.value, 0)
            elif expr.type == "string":
                return expr.value.strip('"')

        # Handling arrays with potential conversion to ctypes arrays
        elif isinstance(expr, c_ast.InitList) and expected_size is not None:
            element_list = [self._parse_expr(e) for e in expr.exprs]

            # Ensure the list matches expected size, filling with zeros if necessary
            if len(element_list) < expected_size:
                element_list.extend([0] * (expected_size - len(element_list)))

            # Convert to ctypes array, dynamically adjusting type based on context if needed
            return (ctypes.c_uint * expected_size)(*element_list)

        # Fallback or default return for unhandled cases
        return None


def merge_structures(structures):
    """Function creating a new type by merging provided types."""
    combined_fields = []
    for struct in structures:
        for field_name, field_type in struct._fields_:
            # Check if the field name already exists
            if field_name not in [field[0] for field in combined_fields]:
                combined_fields.append((field_name, field_type))

    # Create a new structure dynamically
    return type("MergedStructure", (ctypes.Structure,), {'_fields_': combined_fields})


def get_field_type(c, field_name):
    """Function returning the type of a field type based on its name."""
    for field in c._fields_:
        if field[0] == field_name:
            return field[1]

    return None


def create_merged_struct(c, entry):
    """Function returning a fully populate instance of MergedStructure."""
    fields = []
    for field_name, field_type in c._fields_:
        # We might need to 'upgrade' to an array - check field type too.
        # e.g. VIP_SRAM_SIZE_ARRAY -> VIP_SRAM_SIZE_ARRAY[9]

        if hasattr(entry, field_name) and get_field_type(entry, field_name) is field_type:
            fields.append(getattr(entry, field_name))
        else:
            # Add a suitable default value
            if field_type == ctypes.c_uint or field_type == ctypes.c_bool:
                fields.append(0)
            else:
                # It must be an array
                expected_size = getattr(field_type, '_length_')
                fields.append((ctypes.c_uint * expected_size)(*[0] * expected_size))

    return c(*fields)


def enumerate_struct(obj):
    for field_name, field_type in obj._fields_:
        type = 'uint32_t'

        if field_type == ctypes.c_char_p:
            type = 'const char *'

        if field_type == ctypes.c_bool:
            field_name = field_name + ':1'

        if hasattr(field_type, '_length_'):
            field_name = f'{field_name}[{field_type._length_}]'

        yield type, field_name


def enumerate_values(obj):
    for field_name, field_type in obj._fields_:
        value = getattr(obj, field_name)

        if field_type in {ctypes.c_uint, ctypes.c_bool}:
            value = hex(value)
        elif field_type == ctypes.c_char_p:
            value = '"{}"'.format(value.decode('utf-8'))
        elif isinstance(value, ctypes.Array):
            value = '{{{}}}'.format(', '.join(str(element) for element in value))

        yield field_name, value


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--output', required=True, type=str, action="store",
                        help='output C header file')
    parser.add_argument('headers', metavar='header', type=str, nargs='+',
                        help='gc database header to process')
    parser.add_argument('--verbose', "-v", action="store_true",
                        help='be verbose')
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    files = [HeaderFile(header) for header in args.headers]
    structs = [file.database_struct for file in files]
    merged = merge_structures(structs)
    logger.debug('merged struct: %u members', len(merged._fields_))

    entries = []
    for file in files:
        logger.debug('Processing %s', file.filename)
        for entry in file.data:
            merged_entry = create_merged_struct(merged, entry)
            entry_data = list(enumerate_values(merged_entry))
            entries.append(entry_data)

    logger.debug('Total entries: %u', len(entries))

    with open(args.output, "w", encoding="UTF-8") as fh:
        print(Template(template).render(struct=enumerate_struct(merged), entries=entries), file=fh)


if __name__ == '__main__':
    main()



#encoding=utf-8

# Copyright (C) 2016 Intel Corporation
# Copyright (C) 2016 Broadcom
# Copyright (C) 2020 Collabora, Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import textwrap
import xml.etree.ElementTree as ET
import sys

# All instructions in the ISA
instructions = []

MODIFIERS = {}
enums = {}
immediates = []

def xmlbool(s):
    assert(s.lower() in ["false", "true"])
    return False if s.lower() == "false" else True

class EnumValue:
    def __init__(self, value, default):
        self.value = value
        self.default = default

class Enum:
    def __init__(self, name, values):
        self.name = name
        self.values = values
        self.bare_values = [x.value for x in values]

        defaults = [x.value for x in values if x.default]
        if len(defaults) > 0:
            assert(len(defaults) == 1)
            self.default = defaults[0]

def build_enum(el):
    values = []

    for child in el:
        if child.tag == 'value':
            is_default = child.attrib.get('default', False)
            values.append(EnumValue(child.text, is_default))
        elif child.tag == 'reserved':
            values.append(EnumValue("reserved", False))

    return Enum(el.attrib['name'], values)

class Modifier:
    def __init__(self, name, start, size, implied = False, force_enum = None):
        self.name = name
        self.start = start
        self.size = size
        self.implied = implied
        self.is_enum = (force_enum is not None) or size > 1
        self.enum = force_enum or name

        if not self.is_enum:
            self.bare_values = ['', name]
            self.default = 0
        else:
            self.bare_values = [x.value for x in enums[self.enum].values]
            defaults = [x for x in enums[self.enum].values if x.default]
            assert(len(defaults) <= 1)

            if len(defaults) > 0:
                self.default = self.bare_values.index(defaults[0].value)
            else:
                self.default = None

def Flag(name, start):
    return Modifier(name, start, 1)

# Model a single instruction
class Source:
    def __init__(self, index, size, is_float = False, swizzle = False,
            halfswizzle = False, widen = False, lanes = False, combine = False, lane = None, absneg = False, notted = False, name = ""):
        self.is_float = is_float or absneg
        self.start = (index * 8)
        self.size = size
        self.absneg = absneg
        self.notted = notted
        self.swizzle = swizzle
        self.halfswizzle = halfswizzle
        self.widen = widen
        self.lanes = lanes
        self.lane = lane
        self.combine = combine
        self.name = name

        self.offset = {}
        self.bits = {}
        if absneg:
            self.offset['neg'] = 32 + 2 + ((2 - index) * 2)
            self.offset['abs'] = 33 + 2 + ((2 - index) * 2)
            self.bits['neg'] = 1
            self.bits['abs'] = 1
        if notted:
            self.offset['not'] = 35
            self.bits['not'] = 1
        if widen or lanes or halfswizzle:
            self.offset['widen'] = 26 if index == 1 else 36
            self.bits['widen'] = 4 # XXX: too much?
        if lane:
            self.offset['lane'] = self.lane
            self.bits['lane'] = 2 if size in (8, 32) else 1
        if swizzle:
            assert(size in [16, 32])
            self.offset['swizzle'] = 24 + ((2 - index) * 2)
            self.bits['swizzle'] = 2
        if combine:
            self.offset['combine'] = 37
            self.bits['combine'] = 3

class Dest:
    def __init__(self, name = ""):
        self.name = name

class Staging:
    def __init__(self, read = False, write = False, count = 0, flags = 'true', name = ""):
        self.name = name
        self.read = read
        self.write = write
        self.count = count
        self.flags = (flags != 'false')
        self.start = 40

        if write and not self.flags:
            self.start = 16

        # For compatibility
        self.absneg = False
        self.swizzle = False
        self.notted = False
        self.widen = False
        self.lanes = False
        self.lane = False
        self.halfswizzle = False
        self.combine = False
        self.size = 32

        if not self.flags:
            self.encoded_flags = 0
        elif flags == 'rw':
            self.encoded_flags = 0xc0
        else:
            assert(flags == 'true')
            self.encoded_flags = (0x80 if write else 0) | (0x40 if read else 0)

class Immediate:
    def __init__(self, name, start, size, signed):
        self.name = name
        self.start = start
        self.size = size
        self.signed = signed

class Instruction:
    def __init__(self, name, opcode, opcode2, srcs = [], dests = [], immediates = [], modifiers = [], staging = None, unit = None):
        self.name = name
        self.srcs = srcs
        self.dests = dests
        self.opcode = opcode
        self.opcode2 = opcode2 or 0
        self.immediates = immediates
        self.modifiers = modifiers
        self.staging = staging
        self.unit = unit
        self.is_signed = len(name.split(".")) > 1 and ('s' in name.split(".")[1])

        # Message-passing instruction <===> not ALU instruction
        self.message = unit not in ["FMA", "CVT", "SFU"]

        self.secondary_shift = max(len(self.srcs) * 8, 16)
        self.secondary_mask = 0xF if opcode2 is not None else 0x0
        if "left" in [x.name for x in self.modifiers]:
            self.secondary_mask |= 0x100
        if len(srcs) == 3 and (srcs[1].widen or srcs[1].lanes or srcs[1].swizzle):
            self.secondary_mask &= ~0xC # conflicts
        if opcode == 0x90:
            # XXX: XMLify this, but disambiguates sign of conversions
            self.secondary_mask |= 0x10
        if name.startswith("LOAD.i") or name.startswith("STORE.i") or name.startswith("LD_BUFFER.i"):
            self.secondary_shift = 27 # Alias with memory_size
            self.secondary_mask = 0x7
        if "descriptor_type" in [x.name for x in self.modifiers]:
            self.secondary_mask = 0x3
            self.secondary_shift = 37
        elif "memory_width" in [x.name for x in self.modifiers]:
            self.secondary_mask = 0x7
            self.secondary_shift = 27

        assert(len(dests) == 0 or not staging)
        assert(not opcode2 or (opcode2 & self.secondary_mask) == opcode2)

    def __str__(self):
        return self.name

# Build a single source from XML
def build_source(el, i, size):
    lane = el.get('lane', None)
    if lane == "true":
        lane = 38 if i == 0 else 36
    elif lane is not None:
        lane = int(lane)

    return Source(i, int(el.get('size', size)),
            absneg = el.get('absneg', False),
            is_float = el.get('float', False),
            swizzle = el.get('swizzle', False),
            halfswizzle = el.get('halfswizzle', False),
            widen = el.get('widen', False),
            lanes = el.get('lanes', False),
            combine = el.get('combine', False),
            lane = lane,
            notted = el.get('not', False),
            name = el.text or "")

def build_imm(el):
    return Immediate(el.attrib['name'], int(el.attrib['start']),
            int(el.attrib['size']), bool(el.attrib.get('signed', False)))

def build_staging(i, el):
    r = xmlbool(el.attrib.get('read', 'false'))
    w = xmlbool(el.attrib.get('write', 'false'))
    count = int(el.attrib.get('count', '0'))
    flags = el.attrib.get('flags', 'true')

    return Staging(r, w, count, flags, el.text or '')

def build_modifier(el):
    name = el.attrib['name']
    start = int(el.attrib['start'])
    size = int(el.attrib['size'])
    implied = xmlbool(el.get('implied', 'false'))

    return Modifier(name, start, size, implied)

# Build a single instruction from XML and group based overrides
def build_instr(el, overrides = {}):
    # Get overridables
    name = overrides.get('name') or el.attrib.get('name')
    opcode = overrides.get('opcode') or el.attrib.get('opcode')
    opcode2 = overrides.get('opcode2') or el.attrib.get('opcode2')
    unit = overrides.get('unit') or el.attrib.get('unit')
    opcode = int(opcode, base=0)
    opcode2 = int(opcode2, base=0) if opcode2 else None

    # Get explicit sources/dests
    tsize = typesize(name)
    sources = []
    i = 0

    for src in el.findall('src'):
        if (src.attrib.get('pseudo', False)):
            continue
        built = build_source(src, i, tsize)
        sources += [built]

        # 64-bit sources in a 32-bit (message) instruction count as two slots
        # Affects BLEND, ST_CVT
        if tsize != 64 and built.size == 64:
            i = i + 2
        else:
            i = i + 1

    dests = [Dest(dest.text or '') for dest in el.findall('dest')]

    # Get implicit ones
    sources = sources + ([Source(i, int(tsize)) for i in range(int(el.attrib.get('srcs', 0)))])
    dests = dests + ([Dest()] * int(el.attrib.get('dests', 0)))

    # Get staging registers
    staging = [build_staging(i, el) for i, el in enumerate(el.findall('sr'))]

    # Get immediates
    imms = [build_imm(imm) for imm in el.findall('imm')]

    modifiers = []
    for mod in el:
        if (mod.tag in MODIFIERS) and not (mod.attrib.get('pseudo', False)):
            modifiers.append(MODIFIERS[mod.tag])
        elif mod.tag =='va_mod':
            modifiers.append(build_modifier(mod))

    instr = Instruction(name, opcode, opcode2, srcs = sources, dests = dests, immediates = imms, modifiers = modifiers, staging = staging, unit = unit)

    instructions.append(instr)

# Build all the instructions in a group by duplicating the group itself with
# overrides for each distinct instruction
def build_group(el):
    for ins in el.findall('ins'):
        build_instr(el, overrides = {
            'name': ins.attrib['name'],
            'opcode': ins.attrib.get('opcode'),
            'opcode2': ins.attrib.get('opcode2'),
            'unit': ins.attrib.get('unit'),
        })

def to_alphanum(name):
    substitutions = {
        ' ': '_',
        '/': '_',
        '[': '',
        ']': '',
        '(': '',
        ')': '',
        '-': '_',
        ':': '',
        '.': '',
        ',': '',
        '=': '',
        '>': '',
        '#': '',
        '&': '',
        '*': '',
        '"': '',
        '+': '',
        '\'': '',
    }

    for i, j in substitutions.items():
        name = name.replace(i, j)

    return name

def safe_name(name):
    name = to_alphanum(name)
    if not name[0].isalpha():
        name = '_' + name

    return name.lower()

# Parses out the size part of an opcode name
def typesize(opcode):
    if opcode[-3:] == '128':
        return 128
    if opcode[-2:] == '48':
        return 48
    elif opcode[-1] == '8':
        return 8
    else:
        try:
            return int(opcode[-2:])
        except:
            return 32

# Parse the ISA
def valhall_parse_isa(xmlfile = False):
    global MODIFIERS
    global enums
    global immediates
    global root

    xmlfile = os.path.join(os.path.dirname(__file__), 'ISA.xml')
    tree = ET.parse(xmlfile)
    root = tree.getroot()

    # All immediates in the ISA
    ilut = root.findall('lut')[0]
    assert(ilut.attrib['name'] == "Immediates")
    immediates = [int(imm.text, base=0) for imm in ilut.findall('constant')]

    for child in root.findall('enum'):
        enums[safe_name(child.attrib['name'])] = build_enum(child)

    MODIFIERS = {
        # Texture instructions share a common encoding
        "wide_indices": Flag("wide_indices", 8),
        "array_enable": Flag("array_enable", 10),
        "texel_offset": Flag("texel_offset", 11),
        "shadow": Flag("shadow", 12),
        "integer_coordinates": Flag("integer_coordinates", 13),
        "fetch_component": Modifier("fetch_component", 14, 2),
        "lod_mode": Modifier("lod_mode", 13, 3),
        "lod_bias_disable": Modifier("lod_mode", 13, 1),
        "lod_clamp_disable": Modifier("lod_mode", 14, 1),
        "write_mask": Modifier("write_mask", 22, 4),
        "register_type": Modifier("register_type", 26, 2),
        "dimension": Modifier("dimension", 28, 2),
        "skip": Flag("skip", 39),
        "register_width": Modifier("register_width", 46, 1, force_enum = "register_width"),
        "secondary_register_width": Modifier("secondary_register_width", 47, 1, force_enum = "register_width"),
        "vartex_register_width": Modifier("varying_texture_register_width", 24, 2),

        "atom_opc": Modifier("atomic_operation", 22, 4),
        "atom_opc_1": Modifier("atomic_operation_with_1", 22, 4),
        "inactive_result": Modifier("inactive_result", 22, 4),
        "memory_access": Modifier("memory_access", 24, 2),
        "regfmt": Modifier("register_format", 24, 3),
        "source_format": Modifier("source_format", 24, 4),
        "vecsize": Modifier("vector_size", 28, 2),

        "slot": Modifier("slot", 30, 3),
        "roundmode": Modifier("round_mode", 30, 2),
        "result_type": Modifier("result_type", 30, 2),
        "saturate": Flag("saturate", 30),
        "not_result": Flag("not_result", 30),

        "lane_op": Modifier("lane_operation", 32, 2),
        "cmp": Modifier("condition", 32, 3),
        "clamp": Modifier("clamp", 32, 2),
        "sr_count": Modifier("staging_register_count", 33, 3, implied = True),
        "sample_and_update": Modifier("sample_and_update_mode", 33, 3),
        "sr_write_count": Modifier("staging_register_write_count", 36, 3, implied = True),

        "conservative": Flag("conservative", 35),
        "subgroup": Modifier("subgroup_size", 36, 4),
        "update": Modifier("update_mode", 36, 2),
        "sample": Modifier("sample_mode", 38, 2),
    }

    for child in root:
        if child.tag == 'group':
            build_group(child)
        elif child.tag == 'ins':
            build_instr(child)

    instruction_dict = { ins.name: ins for ins in instructions }

    # Validate there are no duplicated instructions
    if len(instruction_dict) != len(instructions):
        import collections
        counts = collections.Counter([i.name for i in instructions])
        for c in counts:
            if counts[c] != 1:
                print(f'{c} appeared {counts[c]} times.')

    assert(len(instruction_dict) == len(instructions))

    return (instructions, immediates, enums, typesize, safe_name)



# Copyright © 2024 Imagination Technologies Ltd.
# SPDX-License-Identifier: MIT

from mako.template import Template, exceptions
from pco_map import *

template = """/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * SPDX-License-Identifier: MIT
 */

#ifndef PCO_MAP_H
#define PCO_MAP_H

/**
 * \\file pco_map.h
 *
 * \\brief PCO mapping definitions.
 */

#include "pco_builder.h"
#include "pco_common.h"
#include "pco_internal.h"
#include "pco_isa.h"
#include "pco_ops.h"
#include "util/macros.h"

#include <stdbool.h>

/* Enum mappings. */
% for enum_map in enum_maps.values():
static inline
${enum_map.type_to} ${enum_map.name}(${enum_map.type_from} val)
{
   uint64_t to = 0;

   % if enum_map.pass_zero_val is not None:
   if (!val)
      return ${enum_map.pass_zero_val};

   % endif
   % if enum_map.both_bitsets:
   u_foreach_bit64 (b, val) {
      switch (b) {
      % for elem_from, elem_to in enum_map.mappings:
      case ${elem_from}:
         to |= ${elem_to};
         break;

      % endfor
      default:
         unreachable("");
      }
   }
   % else:
   switch (val) {
      % for elem_from, elem_to in enum_map.mappings:
   case ${elem_from}:
      to = ${elem_to};
      break;

      % endfor
   default:
      unreachable("");
   }
   % endif

   return to;
}

% endfor
static inline
enum pco_regbank pco_map_reg_bank(pco_ref ref)
{
   enum pco_regbank regbank = pco_map_reg_class_to_regbank(pco_ref_get_reg_class(ref));
   return pco_ref_is_idx_reg(ref) ? regbank + ref.idx_reg.num : regbank;
}

static inline
unsigned pco_map_reg_bank_bits(pco_ref ref)
{
   return util_last_bit(pco_map_reg_bank(ref));
}

static inline
enum pco_idxbank pco_map_idx_bank(pco_ref ref)
{
   assert(pco_ref_is_idx_reg(ref));
   return pco_map_reg_class_to_idxbank(ref.reg_class);
}

static inline unsigned pco_map_reg_index(pco_ref ref)
{
   if (pco_ref_is_reg(ref)) {
      switch (ref.reg_class) {
      case PCO_REG_CLASS_TEMP:
      case PCO_REG_CLASS_VTXIN:
      case PCO_REG_CLASS_COEFF:
      case PCO_REG_CLASS_SHARED:
      case PCO_REG_CLASS_SPEC:
      case PCO_REG_CLASS_CONST:
         return ref.val;

      case PCO_REG_CLASS_INTERN:
         return ref.val + PCO_SR_INTL0;

      case PCO_REG_CLASS_PIXOUT:
         if (ref.val >= 4)
            return ref.val + PCO_SR_PIXOUT4 - 4;
         else
            return ref.val + PCO_SR_PIXOUT0;

      case PCO_REG_CLASS_GLOBAL:
         return ref.val + PCO_SR_GLOBAL0;

      case PCO_REG_CLASS_SLOT:
         return (7 - ref.val) + PCO_SR_SLOT7;

      default:
         break;
      }
   } else if (pco_ref_is_idx_reg(ref)) {
      return pco_map_idx_bank(ref) | (ref.idx_reg.offset << 3);
   }

   unreachable("");
}

static inline unsigned pco_map_reg_index_bits(pco_ref ref)
{
   if (pco_ref_is_reg(ref))
      return util_last_bit(pco_map_reg_index(ref));
   else if (pco_ref_is_idx_reg(ref))
      return 11;

   unreachable("");
}

static inline
enum pco_igrp_hdr_variant pco_igrp_hdr_variant(pco_igrp *igrp)
{
   if (igrp->hdr.alutype == PCO_ALUTYPE_BITWISE)
      return PCO_IGRP_HDR_BITWISE;
   else if (igrp->hdr.alutype == PCO_ALUTYPE_CONTROL)
      return PCO_IGRP_HDR_CONTROL;
   else
      assert(igrp->hdr.alutype == PCO_ALUTYPE_MAIN);

   if (!igrp->hdr.end && !igrp->hdr.atom && !igrp->hdr.rpt && !igrp->hdr.cc)
      return PCO_IGRP_HDR_MAIN_BRIEF;

   return PCO_IGRP_HDR_MAIN;
}

static inline
enum pco_src_variant pco_igrp_src_variant(const pco_igrp *igrp,
                                          bool is_upper)
{
   unsigned offset = is_upper ? ROGUE_ALU_INPUT_GROUP_SIZE : 0;

   pco_ref sA = igrp->srcs.s[0 + offset];
   pco_ref sB = igrp->srcs.s[1 + offset];
   pco_ref sC = igrp->srcs.s[2 + offset];
   pco_ref mux = is_upper ? pco_ref_null() : igrp->iss.is[0];

   bool sA_set = !pco_ref_is_null(sA);
   bool sB_set = !pco_ref_is_null(sB);
   bool sC_set = !pco_ref_is_null(sC);
   bool mux_set = !pco_ref_is_null(mux);

   int sbA_bits = sA_set ? pco_map_reg_bank_bits(sA) : -1;
   int sA_bits = sA_set ? pco_map_reg_index_bits(sA) : -1;

   int sbB_bits = sB_set ? pco_map_reg_bank_bits(sB) : -1;
   int sB_bits = sB_set ? pco_map_reg_index_bits(sB) : -1;

   int sbC_bits = sC_set ? pco_map_reg_bank_bits(sC) : -1;
   int sC_bits = sC_set ? pco_map_reg_index_bits(sC) : -1;

   int mux_bits = mux_set ? util_last_bit(pco_map_io_to_is0_sel(pco_ref_get_io(mux))) : -1;

   bool is_ctrl = igrp->hdr.alutype == PCO_ALUTYPE_CONTROL;
   bool no_srcs_set = !sA_set && !sB_set && !sC_set && !mux_set;
   if (is_ctrl && no_srcs_set)
      return PCO_SRC_NONE;

% for variant, spec in [(bs.name.upper(), bs.data) for bs in I_SRC.bit_structs.values()]:
   ${'else ' if not loop.first else ''}if (${'!' if not spec.is_upper else ''}is_upper &&
            sbA_bits ${'==' if spec.sbA_bits == -1 else '<='} ${spec.sbA_bits} && sA_bits ${'==' if spec.sA_bits == -1 else '<='} ${spec.sA_bits} &&
            sbB_bits ${'==' if spec.sbB_bits == -1 else '<='} ${spec.sbB_bits} && sB_bits ${'==' if spec.sB_bits == -1 else '<='} ${spec.sB_bits} &&
            sbC_bits ${'==' if spec.sbC_bits == -1 else '<='} ${spec.sbC_bits} && sC_bits ${'==' if spec.sC_bits == -1 else '<='} ${spec.sC_bits} &&
            mux_bits ${'==' if spec.mux_bits == -1 else '<='} ${spec.mux_bits}) {
      return ${variant};
   }
% endfor

   unreachable("");
}

static inline
enum pco_iss_variant pco_igrp_iss_variant(const pco_igrp *igrp)
{
   if (igrp->hdr.alutype == PCO_ALUTYPE_MAIN)
      return PCO_ISS_ISS;

   return PCO_ISS_NONE;
}

static inline
enum pco_dst_variant pco_igrp_dest_variant(pco_igrp *igrp)
{
   pco_ref w0 = igrp->dests.w[0];
   pco_ref w1 = igrp->dests.w[1];

   bool w0_set = !pco_ref_is_null(w0);
   bool w1_set = !pco_ref_is_null(w1);

   bool no_dsts = !w0_set && !w1_set;
   bool one_dest = w0_set != w1_set;
   bool dual_dsts = w0_set && w1_set;

   if (no_dsts)
      return PCO_DST_NONE;

   int db0_bits = w0_set ? pco_map_reg_bank_bits(w0) : -1;
   int d0_bits = w0_set ? pco_map_reg_index_bits(w0) : -1;

   int db1_bits = w1_set ? pco_map_reg_bank_bits(w1) : -1;
   int d1_bits = w1_set ? pco_map_reg_index_bits(w1) : -1;

   int dbN_bits = w0_set ? db0_bits : db1_bits;
   int dN_bits = w0_set ? d0_bits : d1_bits;

   if (one_dest) {
      db0_bits = dbN_bits;
      d0_bits = dN_bits;

      db1_bits = -1;
      d1_bits = -1;
   }

% for variant, spec in [(bs.name.upper(), bs.data) for bs in I_DST.bit_structs.values()]:
   ${'else ' if not loop.first else ''}if (${'!' if not spec.dual_dsts else ''}dual_dsts &&
        db0_bits ${'==' if spec.db0_bits == -1 else '<='} ${spec.db0_bits} && d0_bits ${'==' if spec.d0_bits == -1 else '<='} ${spec.d0_bits} &&
        db1_bits ${'==' if spec.db1_bits == -1 else '<='} ${spec.db1_bits} && d1_bits ${'==' if spec.d1_bits == -1 else '<='} ${spec.d1_bits}) {
      return ${variant};
   }
% endfor

   unreachable("");
}

% for encode_map in encode_maps.values():
static inline
unsigned ${encode_map.name}_variant(pco_instr *instr)
{
   % if len(encode_map.variants) > 1:
      % for variant, _, conds in encode_map.variants[1:]:
   ${conds.format('bin', 'instr', 'variant')}
      return ${variant};
      % endfor

   % endif
   return ${encode_map.variants[0][0]};
}

% endfor
/* Instruction group mappings. */
% for group_map in group_maps.values():
static inline
void ${group_map.name}_map_igrp(pco_igrp *igrp, pco_instr *instr)
{
   % for mapping_set in group_map.mapping_sets:
      % for mapping in mapping_set:
   ${mapping.format('igrp', 'instr')}
      % endfor

   % endfor
   igrp->variant.hdr = pco_igrp_hdr_variant(igrp);
   igrp->variant.lower_src = pco_igrp_src_variant(igrp, false);
   igrp->variant.upper_src = pco_igrp_src_variant(igrp, true);
   igrp->variant.iss = pco_igrp_iss_variant(igrp);
   igrp->variant.dest = pco_igrp_dest_variant(igrp);
}

% endfor
static inline
void pco_map_igrp(pco_igrp *igrp, pco_instr *instr)
{
   switch (instr->op) {
% for group_map in group_maps.values():
   case ${group_map.cop_name}:
      return ${group_map.name}_map_igrp(igrp, instr);

% endfor
   default:
      break;
   }

   const struct pco_op_info *info = &pco_op_info[instr->op];
   printf("Instruction group mapping not defined for %s op '%s'.\\n",
          info->type == PCO_OP_TYPE_PSEUDO ? "pseudo" : "hardware",
          info->str);

   unreachable("");
}

static inline unsigned pco_igrp_hdr_map_encode(uint8_t *bin, pco_igrp *igrp)
{
   switch (igrp->variant.hdr) {
   case PCO_IGRP_HDR_MAIN_BRIEF:
      return pco_igrp_hdr_main_brief_encode(bin,
                                            .da = igrp->hdr.da,
                                            .length = igrp->hdr.length,
                                            .oporg = igrp->hdr.oporg,
                                            .olchk = igrp->hdr.olchk,
                                            .w1p = igrp->hdr.w1p,
                                            .w0p = igrp->hdr.w0p,
                                            .cc = igrp->hdr.cc);

   case PCO_IGRP_HDR_MAIN:
      return pco_igrp_hdr_main_encode(bin,
                                      .da = igrp->hdr.da,
                                      .length = igrp->hdr.length,
                                      .oporg = igrp->hdr.oporg,
                                      .olchk = igrp->hdr.olchk,
                                      .w1p = igrp->hdr.w1p,
                                      .w0p = igrp->hdr.w0p,
                                      .cc = igrp->hdr.cc,
                                      .end = igrp->hdr.end,
                                      .atom = igrp->hdr.atom,
                                      .rpt = igrp->hdr.rpt);

   case PCO_IGRP_HDR_BITWISE:
      return pco_igrp_hdr_bitwise_encode(bin,
                                         .da = igrp->hdr.da,
                                         .length = igrp->hdr.length,
                                         .opcnt = igrp->hdr.opcnt,
                                         .olchk = igrp->hdr.olchk,
                                         .w1p = igrp->hdr.w1p,
                                         .w0p = igrp->hdr.w0p,
                                         .cc = igrp->hdr.cc,
                                         .end = igrp->hdr.end,
                                         .atom = igrp->hdr.atom,
                                         .rpt = igrp->hdr.rpt);

   case PCO_IGRP_HDR_CONTROL:
      return pco_igrp_hdr_control_encode(bin,
                                         .da = igrp->hdr.da,
                                         .length = igrp->hdr.length,
                                         .olchk = igrp->hdr.olchk,
                                         .w1p = igrp->hdr.w1p,
                                         .w0p = igrp->hdr.w0p,
                                         .cc = igrp->hdr.cc,
                                         .miscctl = igrp->hdr.miscctl,
                                         .ctrlop = igrp->hdr.ctrlop);

   default:
      break;
   }

   unreachable("");
}

% for encode_map in encode_maps.values():
static inline
   % if len(encode_map.variants) > 1:
unsigned ${encode_map.name}_map_encode(uint8_t *bin, pco_instr *instr, unsigned variant)
{
   switch (variant) {
      % for variant, mapping, _ in encode_map.variants:
   case ${variant}:
      return ${mapping.format('bin', 'instr', 'variant')};

      % endfor
   default:
      break;
   }

   unreachable("");
}
   % else:
unsigned ${encode_map.name}_map_encode(uint8_t *bin, pco_instr *instr)
{
   return ${encode_map.variants[0][1].format('bin', 'instr', 'variant')};
}
   % endif

% endfor
static inline
unsigned pco_instr_map_encode(uint8_t *bin, pco_igrp *igrp, enum pco_op_phase phase)
{
   pco_instr *instr = igrp->instrs[phase];
   switch (instr->op) {
% for encode_map in encode_maps.values():
   case ${encode_map.cop_name}:
   % if len(encode_map.variants) > 1:
      return ${encode_map.name}_map_encode(bin, instr, pco_igrp_variant(igrp, phase));
   % else:
      return ${encode_map.name}_map_encode(bin, instr);
   % endif

% endfor
   default:
      break;
   }

   unreachable("");
}

static inline
unsigned pco_srcs_map_encode(uint8_t *bin, pco_igrp *igrp, bool is_upper)
{
   unsigned offset = is_upper ? ROGUE_ALU_INPUT_GROUP_SIZE : 0;

   pco_ref _sA = igrp->srcs.s[0 + offset];
   pco_ref _sB = igrp->srcs.s[1 + offset];
   pco_ref _sC = igrp->srcs.s[2 + offset];
   pco_ref _mux = is_upper ? pco_ref_null() : igrp->iss.is[0];

   bool sA_set = !pco_ref_is_null(_sA);
   bool sB_set = !pco_ref_is_null(_sB);
   bool sC_set = !pco_ref_is_null(_sC);
   bool mux_set = !pco_ref_is_null(_mux);

   unsigned sbA = sA_set ? pco_map_reg_bank(_sA) : 0;
   unsigned sA = sA_set ? pco_map_reg_index(_sA) : 0;

   unsigned sbB = sB_set ? pco_map_reg_bank(_sB) : 0;
   unsigned sB = sB_set ? pco_map_reg_index(_sB) : 0;

   unsigned sbC = sC_set ? pco_map_reg_bank(_sC) : 0;
   unsigned sC = sC_set ? pco_map_reg_index(_sC) : 0;

   unsigned mux = mux_set ? pco_map_io_to_is0_sel(pco_ref_get_io(_mux)) : 0;

   enum pco_src_variant variant = is_upper ? igrp->variant.upper_src : igrp->variant.lower_src;
   switch (variant) {
% for variant, encode_func, spec, offset in [(bs.name.upper(), f'{bs.name}_encode', bs.data, 3 if bs.data.is_upper else 0) for bs in I_SRC.bit_structs.values()]:
   case ${variant}:
      return ${encode_func}(bin,
   % if spec.sbA_bits != -1:
         .s${offset + 0} = sA,
         .sb${offset + 0} = sbA,
   % endif
   % if spec.sbB_bits != -1:
         .s${offset + 1} = sB,
         .sb${offset + 1} = sbB,
   % endif
   % if spec.sbC_bits != -1:
         .s${offset + 2} = sC,
         .sb${offset + 2} = sbC,
   % endif
   % if spec.mux_bits != -1:
         .is0 = mux,
   % endif
      );

% endfor
   default:
      break;
   }

   unreachable("");
}

static inline
unsigned pco_iss_map_encode(uint8_t *bin, pco_igrp *igrp)
{
   bool is5_set = !pco_ref_is_null(igrp->iss.is[5]);
   bool is4_set = !pco_ref_is_null(igrp->iss.is[4]);
   bool is3_set = !pco_ref_is_null(igrp->iss.is[3]);
   bool is2_set = !pco_ref_is_null(igrp->iss.is[2]);
   bool is1_set = !pco_ref_is_null(igrp->iss.is[1]);

   unsigned is5 = is5_set ? pco_map_io_to_is5_sel(pco_ref_get_io(igrp->iss.is[5])) : 0;
   unsigned is4 = is4_set ? pco_map_io_to_is4_sel(pco_ref_get_io(igrp->iss.is[4])) : 0;
   unsigned is3 = is3_set ? pco_map_io_to_is3_sel(pco_ref_get_io(igrp->iss.is[3])) : 0;
   unsigned is2 = is2_set ? pco_map_io_to_is2_sel(pco_ref_get_io(igrp->iss.is[2])) : 0;
   unsigned is1 = is1_set ? pco_map_io_to_is1_sel(pco_ref_get_io(igrp->iss.is[1])) : 0;

   assert(igrp->variant.iss == PCO_ISS_ISS);
   return pco_iss_iss_encode(bin, .is5 = is5, .is4 = is4, .is3 = is3, .is2 = is2, .is1 = is1);
}

static inline
unsigned pco_dests_map_encode(uint8_t *bin, pco_igrp *igrp)
{
   pco_ref w0 = igrp->dests.w[0];
   pco_ref w1 = igrp->dests.w[1];

   bool w0_set = !pco_ref_is_null(w0);
   bool w1_set = !pco_ref_is_null(w1);

   bool one_dest = w0_set != w1_set;

   int db0 = w0_set ? pco_map_reg_bank(w0) : 0;
   int d0 = w0_set ? pco_map_reg_index(w0) : 0;

   int db1 = w1_set ? pco_map_reg_bank(w1) : 0;
   int d1 = w1_set ? pco_map_reg_index(w1) : 0;

   int dbN = w0_set ? db0 : db1;
   int dN = w0_set ? d0 : d1;

   if (one_dest) {
      db0 = dbN;
      d0 = dN;

      db1 = 0;
      d1 = 0;
   }

   switch (igrp->variant.dest) {
% for variant, encode_func, spec in [(bs.name.upper(), f'{bs.name}_encode', bs.data) for bs in I_DST.bit_structs.values()]:
   case ${variant}:
      return ${encode_func}(bin,
   % if spec.db0_bits != -1:
      % if spec.dual_dsts:
         .d0 = d0,
         .db0 = db0,
      % else:
         .dN = d0,
         .dbN = db0,
      % endif
   % endif
   % if spec.db1_bits != -1:
         .d1 = d1,
         .db1 = db1,
   % endif
      );

% endfor
   default:
      break;
   }

   unreachable("");
}
#endif /* PCO_MAP_H */"""

def main():
   try:
      print(Template(template).render(enum_maps=enum_maps, group_maps=group_maps, encode_maps=encode_maps, I_SRC=I_SRC, I_DST=I_DST))
   except:
       raise Exception(exceptions.text_error_template().render())

if __name__ == '__main__':
   main()



#!/usr/bin/env python3
# Copyright © 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This script reads a meson build directory and gives back the command line it
was configured with.

This only works for meson 0.49.0 and newer.
"""

import argparse
import ast
import configparser
import pathlib
import sys


def parse_args() -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'build_dir',
        help='Path the meson build directory')
    args = parser.parse_args()
    return args


def load_config(path: pathlib.Path) -> configparser.ConfigParser:
    """Load config file."""
    conf = configparser.ConfigParser()
    with path.open() as f:
        conf.read_file(f)
    return conf


def build_cmd(conf: configparser.ConfigParser) -> str:
    """Rebuild the command line."""
    args = []
    for k, v in conf['options'].items():
        if ' ' in v:
            args.append(f'-D{k}="{v}"')
        else:
            args.append(f'-D{k}={v}')

    cf = conf['properties'].get('cross_file')
    if cf:
        args.append('--cross-file={}'.format(cf))
    nf = conf['properties'].get('native_file')
    if nf:
        # this will be in the form "['str', 'str']", so use ast.literal_eval to
        # convert it to a list of strings.
        nf = ast.literal_eval(nf)
        args.extend(['--native-file={}'.format(f) for f in nf])
    return ' '.join(args)


def main():
    args = parse_args()
    path = pathlib.Path(args.build_dir, 'meson-private', 'cmd_line.txt')
    if not path.exists():
        print('Cannot find the necessary file to rebuild command line. '
              'Is your meson version >= 0.49.0?', file=sys.stderr)
        sys.exit(1)

    conf = load_config(path)
    cmd = build_cmd(conf)
    print(cmd)


if __name__ == '__main__':
    main()



COPYRIGHT=u"""
/* Copyright © 2021 Intel Corporation
 * Copyright © 2024 Valve Corporation
 *
 * SPDX-License-Identifier: MIT
 */
"""

import argparse
import os
import sys
import xml.etree.ElementTree as et

import mako
from mako.template import Template

sys.path.append(os.path.join(sys.path[0], '../../../vulkan/util/'))

from vk_entrypoints import get_entrypoints_from_xml

EXCLUDED_COMMANDS = [
    'CmdBeginRenderPass',
    'CmdEndRenderPass',
    'CmdDispatch',
]

TEMPLATE = Template(COPYRIGHT + """
/* This file generated from ${filename}, don't edit directly. */

#include "radv_cmd_buffer.h"
#include "radv_entrypoints.h"

#define ANNOTATE(command, ...) \
   struct radv_cmd_buffer *cmd_buffer = radv_cmd_buffer_from_handle(commandBuffer); \
   struct radv_device *device = radv_cmd_buffer_device(cmd_buffer); \
   radv_cmd_buffer_annotate(cmd_buffer, #command); \
   device->layer_dispatch.annotate.command(__VA_ARGS__)

% for c in commands:
% if c.guard is not None:
#ifdef ${c.guard}
% endif
VKAPI_ATTR ${c.return_type} VKAPI_CALL
annotate_${c.name}(${c.decl_params()})
{
   ANNOTATE(${c.name}, ${c.call_params()});
}

% if c.guard is not None:
#endif // ${c.guard}
% endif
% endfor
""")

# str.removesuffix requires python 3.9+ so implement our own to not break build
# on older versions
def removesuffix(s, suffix):
    l = len(suffix)
    if l == 0:
        return s
    idx = s.find(suffix)
    if idx == len(s) - l:
        return s[:-l]
    return s


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--out-c", required=True, help="Output C file.")
    parser.add_argument("--beta", required=True, help="Enable beta extensions.")
    parser.add_argument("--xml",
                        help="Vulkan API XML file.",
                        required=True, action="append", dest="xml_files")
    args = parser.parse_args()

    commands = []
    commands_names = []
    for e in get_entrypoints_from_xml(args.xml_files, args.beta):
        if not e.name.startswith('Cmd') or e.alias or e.return_type != "void":
            continue

        stripped_name = removesuffix(removesuffix(removesuffix(e.name, 'EXT'), 'KHR'), '2')
        if stripped_name in commands_names or stripped_name in EXCLUDED_COMMANDS:
            continue

        commands.append(e)
        commands_names.append(stripped_name)

    environment = {
        "filename": os.path.basename(__file__),
        "commands": commands,
    }

    try:
        with open(args.out_c, "w", encoding='utf-8') as f:
            f.write(TEMPLATE.render(**environment))
    except Exception:
        # In the event there"s an error, this uses some helpers from mako
        # to print a useful stack trace and prints it, then exits with
        # status 1, if python is run with debug; otherwise it just raises
        # the exception
        print(mako.exceptions.text_error_template().render(), file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()



# Copyright © 2025 Imagination Technologies Ltd.
# SPDX-License-Identifier: MIT

import argparse
import sys
import math

a = 'a'
b = 'b'

lower_algebraic = []
lower_algebraic_late = []

lower_scmp = [
   # Float comparisons + bool conversions.
   (('b2f32', ('flt', a, b)), ('slt', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('fge', a, b)), ('sge', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('feq', a, b)), ('seq', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('fneu', a, b)), ('sne', a, 'b@32'), '!options->lower_scmp'),

   # Float comparisons + bool conversions via bcsel.
   (('bcsel@32', ('flt', a, b), 1.0, 0.0), ('slt', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('fge', a, b), 1.0, 0.0), ('sge', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('feq', a, b), 1.0, 0.0), ('seq', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('fneu', a, b), 1.0, 0.0), ('sne', a, 'b@32'), '!options->lower_scmp'),
]

lower_algebraic_late.extend(lower_scmp)

# TODO: core-specific info.
params=[]

def run():
    import nir_algebraic  # pylint: disable=import-error
    print('#include "pco_internal.h"')
    print(nir_algebraic.AlgebraicPass('pco_nir_lower_algebraic', lower_algebraic, params=params).render())
    print(nir_algebraic.AlgebraicPass('pco_nir_lower_algebraic_late', lower_algebraic_late, params=params).render())

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    args = parser.parse_args()
    sys.path.insert(0, args.import_path)
    run()

if __name__ == '__main__':
    main()



COPYRIGHT = """\
/*
 * Copyright (C) 2017 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
"""

import argparse
import json
from sys import stdout
from mako.template import Template

def collect_data(spirv, kind):
    for x in spirv["operand_kinds"]:
        if x["kind"] == kind:
            operands = x
            break

    values = {}
    for x in operands["enumerants"]:
        val = x["value"]
        assert(val not in values)
        names = [x["enumerant"]]
        if "aliases" in x:
            names.extend(x["aliases"])
        values[val] = names

    return (kind, list(values.values()), operands["category"])

def collect_opcodes(spirv):
    seen = set()
    values = []
    for x in spirv["instructions"]:
        opcode = x["opcode"]
        assert(opcode not in seen)
        seen.add(opcode)

        name = x["opname"]
        assert name.startswith("Op")
        values.append([name[2:]])

    return ("Op", values, None)

def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument('--out-c', required=True, help='Output C file.')
    p.add_argument('--out-h', required=True, help='Output H file.')
    p.add_argument('--json', required=True, help='SPIR-V JSON file.')
    return p.parse_args()

TEMPLATE_H  = Template("""\
/* DO NOT EDIT - This file is generated automatically by spirv_info_c.py script */

""" + COPYRIGHT + """\

#ifndef _SPIRV_INFO_H_
#define _SPIRV_INFO_H_

#include <stdbool.h>

#include "compiler/spirv/spirv.h"

% for kind,values,category in info:
% if kind == "Capability":
struct spirv_capabilities {
    % for names in values:
    % if len(names) == 1:
   bool ${names[0]};
    % else:
   union {
    % for name in names:
      bool ${name};
    % endfor
   };
    % endif
    % endfor
};
% endif
% endfor

bool spirv_capabilities_get(const struct spirv_capabilities *caps,
                            SpvCapability cap);
void spirv_capabilities_set(struct spirv_capabilities *caps,
                            SpvCapability cap, bool enabled);

% for kind,values,category in info:
% if category == "BitEnum":
const char *spirv_${kind.lower()}_to_string(Spv${kind}Mask v);
% else:
const char *spirv_${kind.lower()}_to_string(Spv${kind} v);
% endif
% endfor

#endif /* SPIRV_INFO_H */
""")

TEMPLATE_C  = Template("""\
/* DO NOT EDIT - This file is generated automatically by spirv_info_c.py script */

""" + COPYRIGHT + """\
#include "spirv_info.h"

#include "util/macros.h"

% for kind,values,category in info:
% if kind == "Capability":
bool
spirv_capabilities_get(const struct spirv_capabilities *caps,
                       SpvCapability cap)
{
   switch (cap) {
    % for names in values:
   case SpvCapability${names[0]}: return caps->${names[0]};
    % endfor
   default:
      return false;
   }
}

void
spirv_capabilities_set(struct spirv_capabilities *caps,
                       SpvCapability cap, bool enabled)
{
   switch (cap) {
    % for names in values:
   case SpvCapability${names[0]}: caps->${names[0]} = enabled; break;
    % endfor
   default:
      unreachable("Unknown capability");
   }
}
% endif
% endfor

% for kind,values,category in info:

% if category == "BitEnum":
const char *
spirv_${kind.lower()}_to_string(Spv${kind}Mask v)
{
   switch (v) {
    % for names in values:
    %if names[0] != "None":
   case Spv${kind}${names[0]}Mask: return "Spv${kind}${names[0]}";
    % else:
   case Spv${kind}MaskNone: return "Spv${kind}${names[0]}";
    % endif
    % endfor
   }

   return "unknown";
}
% else:
const char *
spirv_${kind.lower()}_to_string(Spv${kind} v)
{
   switch (v) {
    % for names in values:
   case Spv${kind}${names[0]}: return "Spv${kind}${names[0]}";
    % endfor
   case Spv${kind}Max: break; /* silence warnings about unhandled enums. */
   }

   return "unknown";
}
% endif
% endfor
""")

if __name__ == "__main__":
    pargs = parse_args()

    spirv_info = json.JSONDecoder().decode(open(pargs.json, "r").read())

    info = [
        collect_data(spirv_info, "AddressingModel"),
        collect_data(spirv_info, "BuiltIn"),
        collect_data(spirv_info, "Capability"),
        collect_data(spirv_info, "Decoration"),
        collect_data(spirv_info, "Dim"),
        collect_data(spirv_info, "ExecutionMode"),
        collect_data(spirv_info, "ExecutionModel"),
        collect_data(spirv_info, "FPRoundingMode"),
        collect_data(spirv_info, "FunctionParameterAttribute"),
        collect_data(spirv_info, "ImageFormat"),
        collect_data(spirv_info, "ImageOperands"),
        collect_data(spirv_info, "MemoryModel"),
        collect_data(spirv_info, "StorageClass"),
        collect_opcodes(spirv_info),
    ]

    with open(pargs.out_h, 'w', encoding='utf-8') as f:
        f.write(TEMPLATE_H.render(info=info))
    with open(pargs.out_c, 'w', encoding='utf-8') as f:
        f.write(TEMPLATE_C.render(info=info))



# Copyright 2018 Google LLC
# SPDX-License-Identifier: MIT

from .common.codegen import CodeGen
from .common.vulkantypes import \
        VulkanAPI, makeVulkanTypeSimple, iterateVulkanType, VulkanTypeIterator

from .wrapperdefs import VulkanWrapperGenerator
from .wrapperdefs import STRUCT_EXTENSION_PARAM, STRUCT_EXTENSION_PARAM_FOR_WRITE

class HandleMapCodegen(VulkanTypeIterator):
    def __init__(self, cgen, inputVar, handlemapVarName, prefix, isHandleFunc):
        self.cgen = cgen
        self.inputVar = inputVar
        self.prefix = prefix
        self.handlemapVarName = handlemapVarName

        def makeAccess(varName, asPtr = True):
            return lambda t: self.cgen.generalAccess(t, parentVarName = varName, asPtr = asPtr)

        def makeLengthAccess(varName):
            return lambda t: self.cgen.generalLengthAccess(t, parentVarName = varName)

        def makeLengthAccessGuard(varName):
            return lambda t: self.cgen.generalLengthAccessGuard(t, parentVarName=varName)

        self.exprAccessor = makeAccess(self.inputVar)
        self.exprAccessorValue = makeAccess(self.inputVar, asPtr = False)
        self.lenAccessor = makeLengthAccess(self.inputVar)
        self.lenAccessorGuard = makeLengthAccessGuard(self.inputVar)

        self.checked = False
        self.isHandleFunc = isHandleFunc

    def needSkip(self, vulkanType):
        return False

    def makeCastExpr(self, vulkanType):
        return "(%s)" % (
            self.cgen.makeCTypeDecl(vulkanType, useParamName=False))

    def asNonConstCast(self, access, vulkanType):
        if vulkanType.staticArrExpr:
            casted = "%s(%s)" % (self.makeCastExpr(vulkanType.getForAddressAccess().getForNonConstAccess()), access)
        elif vulkanType.accessibleAsPointer():
            casted = "%s(%s)" % (self.makeCastExpr(vulkanType.getForNonConstAccess()), access)
        else:
            casted = "%s(%s)" % (self.makeCastExpr(vulkanType.getForAddressAccess().getForNonConstAccess()), access)
        return casted

    def onCheck(self, vulkanType):
        pass

    def endCheck(self, vulkanType):
        pass

    def onCompoundType(self, vulkanType):

        if self.needSkip(vulkanType):
            self.cgen.line("// TODO: Unsupported : %s" %
                           self.cgen.makeCTypeDecl(vulkanType))
            return
        access = self.exprAccessor(vulkanType)
        lenAccess = self.lenAccessor(vulkanType)
        lenAccessGuard = self.lenAccessorGuard(vulkanType)

        isPtr = vulkanType.pointerIndirectionLevels > 0

        if lenAccessGuard is not None:
            self.cgen.beginIf(lenAccessGuard)

        if isPtr:
            self.cgen.beginIf(access)

        if lenAccess is not None:

            loopVar = "i"
            access = "%s + %s" % (access, loopVar)
            forInit = "uint32_t %s = 0" % loopVar
            forCond = "%s < (uint32_t)%s" % (loopVar, lenAccess)
            forIncr = "++%s" % loopVar

            self.cgen.beginFor(forInit, forCond, forIncr)
            
        accessCasted = self.asNonConstCast(access, vulkanType)
        self.cgen.funcCall(None, self.prefix + vulkanType.typeName,
                           [self.handlemapVarName, accessCasted])

        if lenAccess is not None:
            self.cgen.endFor()

        if isPtr:
            self.cgen.endIf()

        if lenAccessGuard is not None:
            self.cgen.endIf()

    def onString(self, vulkanType):
        pass

    def onStringArray(self, vulkanType):
        pass

    def onStaticArr(self, vulkanType):
        if not self.isHandleFunc(vulkanType):
            return

        accessLhs = self.exprAccessor(vulkanType)
        lenAccess = self.lenAccessor(vulkanType)

        self.cgen.stmt("%s->mapHandles_%s(%s%s, %s)" % \
            (self.handlemapVarName, vulkanType.typeName,
             self.makeCastExpr(vulkanType.getForAddressAccess().getForNonConstAccess()),
             accessLhs, lenAccess))

    def onStructExtension(self, vulkanType):
        access = self.exprAccessor(vulkanType)

        castedAccessExpr = "(%s)(%s)" % ("void*", access)
        self.cgen.beginIf(access)
        self.cgen.funcCall(None, self.prefix + "extension_struct",
                           [self.handlemapVarName, castedAccessExpr])
        self.cgen.endIf()

    def onPointer(self, vulkanType):
        if self.needSkip(vulkanType):
            return

        if not self.isHandleFunc(vulkanType):
            return

        access = self.exprAccessor(vulkanType)
        lenAccess = self.lenAccessor(vulkanType)
        lenAccess = "1" if lenAccess is None else lenAccess

        self.cgen.beginIf(access)

        self.cgen.stmt( \
            "%s->mapHandles_%s(%s%s, %s)" % \
            (self.handlemapVarName,
             vulkanType.typeName,
             self.makeCastExpr(vulkanType.getForNonConstAccess()),
             access,
             lenAccess))

        self.cgen.endIf()

    def onValue(self, vulkanType):
        if not self.isHandleFunc(vulkanType):
            return
        access = self.exprAccessor(vulkanType)
        self.cgen.stmt(
            "%s->mapHandles_%s(%s%s)" % \
            (self.handlemapVarName, vulkanType.typeName,
             self.makeCastExpr(vulkanType.getForAddressAccess().getForNonConstAccess()),
             access))

class VulkanHandleMap(VulkanWrapperGenerator):
    def __init__(self, module, typeInfo):
        VulkanWrapperGenerator.__init__(self, module, typeInfo)

        self.codegen = CodeGen()

        self.handlemapPrefix = "handlemap_"
        self.toMapVar = "toMap"
        self.handlemapVarName = "handlemap"
        self.handlemapParam = \
            makeVulkanTypeSimple(False, "VulkanHandleMapping", 1,
                                 self.handlemapVarName)
        self.voidType = makeVulkanTypeSimple(False, "void", 0)

        self.handlemapCodegen = \
            HandleMapCodegen(
                None,
                self.toMapVar,
                self.handlemapVarName,
                self.handlemapPrefix,
                lambda vtype : typeInfo.isHandleType(vtype.typeName))

        self.knownDefs = {}

        self.extensionHandlemapPrototype = \
            VulkanAPI(self.handlemapPrefix + "extension_struct",
                      self.voidType,
                      [self.handlemapParam, STRUCT_EXTENSION_PARAM_FOR_WRITE])

    def onBegin(self,):
        VulkanWrapperGenerator.onBegin(self)
        self.module.appendImpl(self.codegen.makeFuncDecl(
            self.extensionHandlemapPrototype))

    def onGenType(self, typeXml, name, alias):
        VulkanWrapperGenerator.onGenType(self, typeXml, name, alias)

        if name in self.knownDefs:
            return

        category = self.typeInfo.categoryOf(name)

        if category in ["struct", "union"] and alias:
            self.module.appendHeader(
                self.codegen.makeFuncAlias(self.handlemapPrefix + name,
                                           self.handlemapPrefix + alias))

        if category in ["struct", "union"] and not alias:

            structInfo = self.typeInfo.structs[name]

            typeFromName = \
                lambda varname: \
                    makeVulkanTypeSimple(varname == "from", name, 1, varname)

            handlemapParams = \
                [self.handlemapParam] + \
                list(map(typeFromName, [self.toMapVar]))

            handlemapPrototype = \
                VulkanAPI(self.handlemapPrefix + name,
                          self.voidType,
                          handlemapParams)

            def funcDefGenerator(cgen):
                self.handlemapCodegen.cgen = cgen
                for p in handlemapParams:
                    cgen.stmt("(void)%s" % p.paramName)
                for member in structInfo.members:
                    iterateVulkanType(self.typeInfo, member,
                                      self.handlemapCodegen)

            self.module.appendHeader(
                self.codegen.makeFuncDecl(handlemapPrototype))
            self.module.appendImpl(
                self.codegen.makeFuncImpl(handlemapPrototype, funcDefGenerator))

    def onGenCmd(self, cmdinfo, name, alias):
        VulkanWrapperGenerator.onGenCmd(self, cmdinfo, name, alias)

    def onEnd(self,):
        VulkanWrapperGenerator.onEnd(self)

        def forEachExtensionHandlemap(ext, castedAccess, cgen):
            cgen.funcCall(None, self.handlemapPrefix + ext.name,
                          [self.handlemapVarName, castedAccess])

        self.module.appendImpl(
            self.codegen.makeFuncImpl(
                self.extensionHandlemapPrototype,
                lambda cgen: self.emitForEachStructExtension(
                    cgen,
                    self.voidType,
                    STRUCT_EXTENSION_PARAM_FOR_WRITE,
                    forEachExtensionHandlemap)))



#
# Copyright (C) 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import ast
from collections import defaultdict
import itertools
import struct
import sys
import mako.template
import re
import traceback

from nir_opcodes import opcodes, type_sizes

# This should be the same as NIR_SEARCH_MAX_COMM_OPS in nir_search.c
nir_search_max_comm_ops = 8

# These opcodes are only employed by nir_search.  This provides a mapping from
# opcode to destination type.
conv_opcode_types = {
    'i2f' : 'float',
    'u2f' : 'float',
    'f2f' : 'float',
    'f2u' : 'uint',
    'f2i' : 'int',
    'u2u' : 'uint',
    'i2i' : 'int',
    'b2f' : 'float',
    'b2i' : 'int',
    'i2b' : 'bool',
    'f2b' : 'bool',
}

swizzles = {'x' : 0, 'y' : 1, 'z' : 2, 'w' : 3,
            'a' : 0, 'b' : 1, 'c' : 2, 'd' : 3,
            'e' : 4, 'f' : 5, 'g' : 6, 'h' : 7,
            'i' : 8, 'j' : 9, 'k' : 10, 'l' : 11,
            'm' : 12, 'n' : 13, 'o' : 14, 'p' : 15 }

def get_cond_index(conds, cond):
    if cond:
        if cond in conds:
            return conds[cond]
        else:
            cond_index = len(conds)
            conds[cond] = cond_index
            return cond_index
    else:
        return -1

def get_c_opcode(op):
      if op in conv_opcode_types:
         return 'nir_search_op_' + op
      else:
         return 'nir_op_' + op

_type_re = re.compile(r"(?P<type>int|uint|bool|float)?(?P<bits>\d+)?")

def type_bits(type_str):
   m = _type_re.match(type_str)
   assert m.group('type')

   if m.group('bits') is None:
      return 0
   else:
      return int(m.group('bits'))

# Represents a set of variables, each with a unique id
class VarSet(object):
   def __init__(self):
      self.names = {}
      self.ids = itertools.count()
      self.immutable = False;

   def __getitem__(self, name):
      if name not in self.names:
         assert not self.immutable, "Unknown replacement variable: " + name
         self.names[name] = next(self.ids)

      return self.names[name]

   def lock(self):
      self.immutable = True

class SearchExpression(object):
   def __init__(self, expr):
      self.opcode = expr[0]
      self.sources = expr[1:]
      self.ignore_exact = False

   @staticmethod
   def create(val):
      if isinstance(val, tuple):
         return SearchExpression(val)
      else:
         assert(isinstance(val, SearchExpression))
         return val

   def __repr__(self):
      l = [self.opcode, *self.sources]
      if self.ignore_exact:
         l.append('ignore_exact')
      return repr((*l,))

class Value(object):
   @staticmethod
   def create(val, name_base, varset, algebraic_pass):
      if isinstance(val, bytes):
         val = val.decode('utf-8')

      if isinstance(val, tuple) or isinstance(val, SearchExpression):
         return Expression(val, name_base, varset, algebraic_pass)
      elif isinstance(val, Expression):
         return val
      elif isinstance(val, str):
         return Variable(val, name_base, varset, algebraic_pass)
      elif isinstance(val, (bool, float, int)):
         return Constant(val, name_base)

   def __init__(self, val, name, type_str):
      self.in_val = str(val)
      self.name = name
      self.type_str = type_str

   def __str__(self):
      return self.in_val

   def get_bit_size(self):
      """Get the physical bit-size that has been chosen for this value, or if
      there is none, the canonical value which currently represents this
      bit-size class. Variables will be preferred, i.e. if there are any
      variables in the equivalence class, the canonical value will be a
      variable. We do this since we'll need to know which variable each value
      is equivalent to when constructing the replacement expression. This is
      the "find" part of the union-find algorithm.
      """
      bit_size = self

      while isinstance(bit_size, Value):
         if bit_size._bit_size is None:
            break
         bit_size = bit_size._bit_size

      if bit_size is not self:
         self._bit_size = bit_size
      return bit_size

   def set_bit_size(self, other):
      """Make self.get_bit_size() return what other.get_bit_size() return
      before calling this, or just "other" if it's a concrete bit-size. This is
      the "union" part of the union-find algorithm.
      """

      self_bit_size = self.get_bit_size()
      other_bit_size = other if isinstance(other, int) else other.get_bit_size()

      if self_bit_size == other_bit_size:
         return

      self_bit_size._bit_size = other_bit_size

   @property
   def type_enum(self):
      return "nir_search_value_" + self.type_str

   @property
   def c_bit_size(self):
      bit_size = self.get_bit_size()
      if isinstance(bit_size, int):
         return bit_size
      elif isinstance(bit_size, Variable):
         return -bit_size.index - 1
      else:
         # If the bit-size class is neither a variable, nor an actual bit-size, then
         # - If it's in the search expression, we don't need to check anything
         # - If it's in the replace expression, either it's ambiguous (in which
         # case we'd reject it), or it equals the bit-size of the search value
         # We represent these cases with a 0 bit-size.
         return 0

   __template = mako.template.Template("""   { .${val.type_str} = {
      { ${val.type_enum}, ${val.c_bit_size} },
% if isinstance(val, Constant):
      ${val.type()}, { ${val.hex()} /* ${val.value} */ },
% elif isinstance(val, Variable):
      ${val.index}, /* ${val.var_name} */
      ${'true' if val.is_constant else 'false'},
      ${val.type() or 'nir_type_invalid' },
      ${val.cond_index},
      ${val.swizzle()},
% elif isinstance(val, Expression):
      ${'true' if val.inexact else 'false'},
      ${'true' if val.exact else 'false'},
      ${'true' if val.ignore_exact else 'false'},
      ${'true' if val.nsz else 'false'},
      ${'true' if val.nnan else 'false'},
      ${'true' if val.ninf else 'false'},
      ${'true' if val.contract else 'false'},
      ${val.swizzle},
      ${val.c_opcode()},
      ${val.comm_expr_idx}, ${val.comm_exprs},
      { ${', '.join(src.array_index for src in val.sources)} },
      ${val.cond_index},
% endif
   } },
""")

   def render(self, cache):
      struct_init = self.__template.render(val=self,
                                           Constant=Constant,
                                           Variable=Variable,
                                           Expression=Expression)
      if struct_init in cache:
         # If it's in the cache, register a name remap in the cache and render
         # only a comment saying it's been remapped
         self.array_index = cache[struct_init]
         return "   /* {} -> {} in the cache */\n".format(self.name,
                                                       cache[struct_init])
      else:
         self.array_index = str(cache["next_index"])
         cache[struct_init] = self.array_index
         cache["next_index"] += 1
         return struct_init

_constant_re = re.compile(r"(?P<value>[^@\(]+)(?:@(?P<bits>\d+))?")

class Constant(Value):
   def __init__(self, val, name):
      Value.__init__(self, val, name, "constant")

      if isinstance(val, (str)):
         m = _constant_re.match(val)
         self.value = ast.literal_eval(m.group('value'))
         self._bit_size = int(m.group('bits')) if m.group('bits') else None
      else:
         self.value = val
         self._bit_size = None

      if isinstance(self.value, bool):
         assert self._bit_size is None or self._bit_size == 1
         self._bit_size = 1

   def hex(self):
      if isinstance(self.value, (bool)):
         return 'NIR_TRUE' if self.value else 'NIR_FALSE'
      if isinstance(self.value, int):
         # Explicitly sign-extend negative integers to 64-bit, ensuring correct
         # handling of -INT32_MIN which is not representable in 32-bit.
         if self.value < 0:
            return hex(struct.unpack('Q', struct.pack('q', self.value))[0]) + 'ull'
         else:
            return hex(self.value) + 'ull'
      elif isinstance(self.value, float):
         return hex(struct.unpack('Q', struct.pack('d', self.value))[0]) + 'ull'
      else:
         assert False

   def type(self):
      if isinstance(self.value, (bool)):
         return "nir_type_bool"
      elif isinstance(self.value, int):
         return "nir_type_int"
      elif isinstance(self.value, float):
         return "nir_type_float"

   def equivalent(self, other):
      """Check that two constants are equivalent.

      This is check is much weaker than equality.  One generally cannot be
      used in place of the other.  Using this implementation for the __eq__
      will break BitSizeValidator.

      """
      if not isinstance(other, type(self)):
         return False

      return self.value == other.value

# The $ at the end forces there to be an error if any part of the string
# doesn't match one of the field patterns.
_var_name_re = re.compile(r"(?P<const>#)?(?P<name>\w+)"
                          r"(?:@(?P<type>int|uint|bool|float)?(?P<bits>\d+)?)?"
                          r"(?P<cond>\([^\)]+\))?"
                          r"(?P<swiz>\.[xyzwabcdefghijklmnop]+)?"
                          r"$")

class Variable(Value):
   def __init__(self, val, name, varset, algebraic_pass):
      Value.__init__(self, val, name, "variable")

      m = _var_name_re.match(val)
      assert m and m.group('name') is not None, \
            "Malformed variable name \"{}\".".format(val)

      self.var_name = m.group('name')

      # Prevent common cases where someone puts quotes around a literal
      # constant.  If we want to support names that have numeric or
      # punctuation characters, we can me the first assertion more flexible.
      assert self.var_name.isalpha()
      assert self.var_name != 'True'
      assert self.var_name != 'False'

      self.is_constant = m.group('const') is not None
      self.cond_index = get_cond_index(algebraic_pass.variable_cond, m.group('cond'))
      self.required_type = m.group('type')
      self._bit_size = int(m.group('bits')) if m.group('bits') else None
      self.swiz = m.group('swiz')

      if self.required_type == 'bool':
         if self._bit_size is not None:
            assert self._bit_size in type_sizes(self.required_type)
         else:
            self._bit_size = 1

      if self.required_type is not None:
         assert self.required_type in ('float', 'bool', 'int', 'uint')

      self.index = varset[self.var_name]

   def type(self):
      if self.required_type == 'bool':
         return "nir_type_bool"
      elif self.required_type in ('int', 'uint'):
         return "nir_type_int"
      elif self.required_type == 'float':
         return "nir_type_float"

   def equivalent(self, other):
      """Check that two variables are equivalent.

      This is check is much weaker than equality.  One generally cannot be
      used in place of the other.  Using this implementation for the __eq__
      will break BitSizeValidator.

      """
      if not isinstance(other, type(self)):
         return False

      return self.index == other.index

   def swizzle(self):
      if self.swiz is not None:
         return '{' + ', '.join([str(swizzles[c]) for c in self.swiz[1:]]) + '}'
      return '{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}'

_opcode_re = re.compile(r"(?P<inexact>~)?(?P<exact>!)?(?P<opcode>\w+)(?:@(?P<bits>\d+))?"
                        r"(?P<cond>\([^\)]+\))?(?P<swizzle>\.[xyzwabcdefghijklmnop])?")

class Expression(Value):
   def __init__(self, expr, name_base, varset, algebraic_pass):
      Value.__init__(self, expr, name_base, "expression")

      expr = SearchExpression.create(expr)

      m = _opcode_re.match(expr.opcode)
      assert m and m.group('opcode') is not None

      self.opcode = m.group('opcode')
      self._bit_size = int(m.group('bits')) if m.group('bits') else None
      self.inexact = m.group('inexact') is not None
      self.exact = m.group('exact') is not None
      self.ignore_exact = expr.ignore_exact
      self.cond = m.group('cond')

      assert not self.inexact or not self.exact, \
            'Expression cannot be both exact and inexact.'

      # "many-comm-expr" isn't really a condition.  It's notification to the
      # generator that this pattern is known to have too many commutative
      # expressions, and an error should not be generated for this case.
      # nsz, nnan and ninf are special conditions, so we treat them specially too.
      cond = {k: True for k in self.cond[1:-1].split(",")} if self.cond else {}
      self.many_commutative_expressions = cond.pop('many-comm-expr', False)
      self.nsz = cond.pop('nsz', False)
      self.nnan = cond.pop('nnan', False)
      self.ninf = cond.pop('ninf', False)
      self.contract = cond.pop('contract', False)
      self.swizzle = -1 if m.group('swizzle') is None else swizzles[m.group('swizzle').removeprefix('.')]

      assert len(cond) <= 1
      self.cond = cond.popitem()[0] if cond else None

      # Deduplicate references to the condition functions for the expressions
      # and save the index for the order they were added.
      self.cond_index = get_cond_index(algebraic_pass.expression_cond, self.cond)

      self.sources = [ Value.create(src, "{0}_{1}".format(name_base, i), varset, algebraic_pass)
                       for (i, src) in enumerate(expr.sources) ]

      # nir_search_expression::srcs is hard-coded to 4
      assert len(self.sources) <= 4

      if self.opcode in conv_opcode_types:
         assert self._bit_size is None, \
                'Expression cannot use an unsized conversion opcode with ' \
                'an explicit size; that\'s silly.'

      self.__index_comm_exprs(0)

   def equivalent(self, other):
      """Check that two variables are equivalent.

      This is check is much weaker than equality.  One generally cannot be
      used in place of the other.  Using this implementation for the __eq__
      will break BitSizeValidator.

      This implementation does not check for equivalence due to commutativity,
      but it could.

      """
      if not isinstance(other, type(self)):
         return False

      if len(self.sources) != len(other.sources):
         return False

      if self.opcode != other.opcode:
         return False

      return all(s.equivalent(o) for s, o in zip(self.sources, other.sources))

   def __index_comm_exprs(self, base_idx):
      """Recursively count and index commutative expressions
      """
      self.comm_exprs = 0

      # A note about the explicit "len(self.sources)" check. The list of
      # sources comes from user input, and that input might be bad.  Check
      # that the expected second source exists before accessing it. Without
      # this check, a unit test that does "('iadd', 'a')" will crash.
      if self.opcode not in conv_opcode_types and \
         "2src_commutative" in opcodes[self.opcode].algebraic_properties and \
         len(self.sources) >= 2 and \
         not self.sources[0].equivalent(self.sources[1]):
         self.comm_expr_idx = base_idx
         self.comm_exprs += 1
      else:
         self.comm_expr_idx = -1

      for s in self.sources:
         if isinstance(s, Expression):
            s.__index_comm_exprs(base_idx + self.comm_exprs)
            self.comm_exprs += s.comm_exprs

      return self.comm_exprs

   def c_opcode(self):
      return get_c_opcode(self.opcode)

   def render(self, cache):
      srcs = "".join(src.render(cache) for src in self.sources)
      return srcs + super(Expression, self).render(cache)

class BitSizeValidator(object):
   """A class for validating bit sizes of expressions.

   NIR supports multiple bit-sizes on expressions in order to handle things
   such as fp64.  The source and destination of every ALU operation is
   assigned a type and that type may or may not specify a bit size.  Sources
   and destinations whose type does not specify a bit size are considered
   "unsized" and automatically take on the bit size of the corresponding
   register or SSA value.  NIR has two simple rules for bit sizes that are
   validated by nir_validator:

    1) A given SSA def or register has a single bit size that is respected by
       everything that reads from it or writes to it.

    2) The bit sizes of all unsized inputs/outputs on any given ALU
       instruction must match.  They need not match the sized inputs or
       outputs but they must match each other.

   In order to keep nir_algebraic relatively simple and easy-to-use,
   nir_search supports a type of bit-size inference based on the two rules
   above.  This is similar to type inference in many common programming
   languages.  If, for instance, you are constructing an add operation and you
   know the second source is 16-bit, then you know that the other source and
   the destination must also be 16-bit.  There are, however, cases where this
   inference can be ambiguous or contradictory.  Consider, for instance, the
   following transformation:

   (('usub_borrow', a, b), ('b2i@32', ('ult', a, b)))

   This transformation can potentially cause a problem because usub_borrow is
   well-defined for any bit-size of integer.  However, b2i always generates a
   32-bit result so it could end up replacing a 64-bit expression with one
   that takes two 64-bit values and produces a 32-bit value.  As another
   example, consider this expression:

   (('bcsel', a, b, 0), ('iand', a, b))

   In this case, in the search expression a must be 32-bit but b can
   potentially have any bit size.  If we had a 64-bit b value, we would end up
   trying to and a 32-bit value with a 64-bit value which would be invalid

   This class solves that problem by providing a validation layer that proves
   that a given search-and-replace operation is 100% well-defined before we
   generate any code.  This ensures that bugs are caught at compile time
   rather than at run time.

   Each value maintains a "bit-size class", which is either an actual bit size
   or an equivalence class with other values that must have the same bit size.
   The validator works by combining bit-size classes with each other according
   to the NIR rules outlined above, checking that there are no inconsistencies.
   When doing this for the replacement expression, we make sure to never change
   the equivalence class of any of the search values. We could make the example
   transforms above work by doing some extra run-time checking of the search
   expression, but we make the user specify those constraints themselves, to
   avoid any surprises. Since the replacement bitsizes can only be connected to
   the source bitsize via variables (variables must have the same bitsize in
   the source and replacment expressions) or the roots of the expression (the
   replacement expression must produce the same bit size as the search
   expression), we prevent merging a variable with anything when processing the
   replacement expression, or specializing the search bitsize
   with anything. The former prevents

   (('bcsel', a, b, 0), ('iand', a, b))

   from being allowed, since we'd have to merge the bitsizes for a and b due to
   the 'iand', while the latter prevents

   (('usub_borrow', a, b), ('b2i@32', ('ult', a, b)))

   from being allowed, since the search expression has the bit size of a and b,
   which can't be specialized to 32 which is the bitsize of the replace
   expression. It also prevents something like:

   (('b2i', ('i2b', a)), ('ineq', a, 0))

   since the bitsize of 'b2i', which can be anything, can't be specialized to
   the bitsize of a.

   After doing all this, we check that every subexpression of the replacement
   was assigned a constant bitsize, the bitsize of a variable, or the bitsize
   of the search expresssion, since those are the things that are known when
   constructing the replacement expresssion. Finally, we record the bitsize
   needed in nir_search_value so that we know what to do when building the
   replacement expression.
   """

   def __init__(self, varset):
      self._var_classes = [None] * len(varset.names)

   def compare_bitsizes(self, a, b):
      """Determines which bitsize class is a specialization of the other, or
      whether neither is. When we merge two different bitsizes, the
      less-specialized bitsize always points to the more-specialized one, so
      that calling get_bit_size() always gets you the most specialized bitsize.
      The specialization partial order is given by:
      - Physical bitsizes are always the most specialized, and a different
        bitsize can never specialize another.
      - In the search expression, variables can always be specialized to each
        other and to physical bitsizes. In the replace expression, we disallow
        this to avoid adding extra constraints to the search expression that
        the user didn't specify.
      - Expressions and constants without a bitsize can always be specialized to
        each other and variables, but not the other way around.

        We return -1 if a <= b (b can be specialized to a), 0 if a = b, 1 if a >= b,
        and None if they are not comparable (neither a <= b nor b <= a).
      """
      if isinstance(a, int):
         if isinstance(b, int):
            return 0 if a == b else None
         elif isinstance(b, Variable):
            return -1 if self.is_search else None
         else:
            return -1
      elif isinstance(a, Variable):
         if isinstance(b, int):
            return 1 if self.is_search else None
         elif isinstance(b, Variable):
            return 0 if self.is_search or a.index == b.index else None
         else:
            return -1
      else:
         if isinstance(b, int):
            return 1
         elif isinstance(b, Variable):
            return 1
         else:
            return 0

   def unify_bit_size(self, a, b, error_msg):
      """Record that a must have the same bit-size as b. If both
      have been assigned conflicting physical bit-sizes, call "error_msg" with
      the bit-sizes of self and other to get a message and raise an error.
      In the replace expression, disallow merging variables with other
      variables and physical bit-sizes as well.
      """
      a_bit_size = a.get_bit_size()
      b_bit_size = b if isinstance(b, int) else b.get_bit_size()

      cmp_result = self.compare_bitsizes(a_bit_size, b_bit_size)

      assert cmp_result is not None, \
         error_msg(a_bit_size, b_bit_size)

      if cmp_result < 0:
         b_bit_size.set_bit_size(a)
      elif not isinstance(a_bit_size, int):
         a_bit_size.set_bit_size(b)

   def merge_variables(self, val):
      """Perform the first part of type inference by merging all the different
      uses of the same variable. We always do this as if we're in the search
      expression, even if we're actually not, since otherwise we'd get errors
      if the search expression specified some constraint but the replace
      expression didn't, because we'd be merging a variable and a constant.
      """
      if isinstance(val, Variable):
         if self._var_classes[val.index] is None:
            self._var_classes[val.index] = val
         else:
            other = self._var_classes[val.index]
            self.unify_bit_size(other, val,
                  lambda other_bit_size, bit_size:
                     'Variable {} has conflicting bit size requirements: ' \
                     'it must have bit size {} and {}'.format(
                        val.var_name, other_bit_size, bit_size))
      elif isinstance(val, Expression):
         for src in val.sources:
            self.merge_variables(src)

   def validate_value(self, val):
      """Validate the an expression by performing classic Hindley-Milner
      type inference on bitsizes. This will detect if there are any conflicting
      requirements, and unify variables so that we know which variables must
      have the same bitsize. If we're operating on the replace expression, we
      will refuse to merge different variables together or merge a variable
      with a constant, in order to prevent surprises due to rules unexpectedly
      not matching at runtime.
      """
      if not isinstance(val, Expression):
         return

      # Generic conversion ops are special in that they have a single unsized
      # source and an unsized destination and the two don't have to match.
      # This means there's no validation or unioning to do here besides the
      # len(val.sources) check.
      if val.opcode in conv_opcode_types:
         assert len(val.sources) == 1, \
            "Expression {} has {} sources, expected 1".format(
               val, len(val.sources))
         self.validate_value(val.sources[0])
         return

      nir_op = opcodes[val.opcode]
      assert len(val.sources) == nir_op.num_inputs, \
         "Expression {} has {} sources, expected {}".format(
            val, len(val.sources), nir_op.num_inputs)

      for src in val.sources:
         self.validate_value(src)

      dst_type_bits = type_bits(nir_op.output_type)

      # First, unify all the sources. That way, an error coming up because two
      # sources have an incompatible bit-size won't produce an error message
      # involving the destination.
      first_unsized_src = None
      for src_type, src in zip(nir_op.input_types, val.sources):
         src_type_bits = type_bits(src_type)
         if src_type_bits == 0:
            if first_unsized_src is None:
               first_unsized_src = src
               continue

            if self.is_search:
               self.unify_bit_size(first_unsized_src, src,
                  lambda first_unsized_src_bit_size, src_bit_size:
                     'Source {} of {} must have bit size {}, while source {} ' \
                     'must have incompatible bit size {}'.format(
                        first_unsized_src, val, first_unsized_src_bit_size,
                        src, src_bit_size))
            else:
               self.unify_bit_size(first_unsized_src, src,
                  lambda first_unsized_src_bit_size, src_bit_size:
                     'Sources {} (bit size of {}) and {} (bit size of {}) ' \
                     'of {} may not have the same bit size when building the ' \
                     'replacement expression.'.format(
                        first_unsized_src, first_unsized_src_bit_size, src,
                        src_bit_size, val))
         else:
            if self.is_search:
               self.unify_bit_size(src, src_type_bits,
                  lambda src_bit_size, unused:
                     '{} must have {} bits, but as a source of nir_op_{} '\
                     'it must have {} bits'.format(
                        src, src_bit_size, nir_op.name, src_type_bits))
            else:
               self.unify_bit_size(src, src_type_bits,
                  lambda src_bit_size, unused:
                     '{} has the bit size of {}, but as a source of ' \
                     'nir_op_{} it must have {} bits, which may not be the ' \
                     'same'.format(
                        src, src_bit_size, nir_op.name, src_type_bits))

      if dst_type_bits == 0:
         if first_unsized_src is not None:
            if self.is_search:
               self.unify_bit_size(val, first_unsized_src,
                  lambda val_bit_size, src_bit_size:
                     '{} must have the bit size of {}, while its source {} ' \
                     'must have incompatible bit size {}'.format(
                        val, val_bit_size, first_unsized_src, src_bit_size))
            else:
               self.unify_bit_size(val, first_unsized_src,
                  lambda val_bit_size, src_bit_size:
                     '{} must have {} bits, but its source {} ' \
                     '(bit size of {}) may not have that bit size ' \
                     'when building the replacement.'.format(
                        val, val_bit_size, first_unsized_src, src_bit_size))
      else:
         self.unify_bit_size(val, dst_type_bits,
            lambda dst_bit_size, unused:
               '{} must have {} bits, but as a destination of nir_op_{} ' \
               'it must have {} bits'.format(
                  val, dst_bit_size, nir_op.name, dst_type_bits))

   def validate_replace(self, val, search):
      bit_size = val.get_bit_size()
      assert isinstance(bit_size, int) or isinstance(bit_size, Variable) or \
            bit_size == search.get_bit_size(), \
            'Ambiguous bit size for replacement value {}: ' \
            'it cannot be deduced from a variable, a fixed bit size ' \
            'somewhere, or the search expression.'.format(val)

      if isinstance(val, Expression):
         for src in val.sources:
            self.validate_replace(src, search)
      elif isinstance(val, Variable):
          # These catch problems when someone copies and pastes the search
          # into the replacement.
          assert not val.is_constant, \
              'Replacement variables must not be marked constant.'

          assert val.cond_index == -1, \
              'Replacement variables must not have a condition.'

          assert not val.required_type, \
              'Replacement variables must not have a required type.'

   def validate(self, search, replace):
      self.is_search = True
      self.merge_variables(search)
      self.merge_variables(replace)
      self.validate_value(search)

      self.is_search = False
      self.validate_value(replace)

      # Check that search is always more specialized than replace. Note that
      # we're doing this in replace mode, disallowing merging variables.
      search_bit_size = search.get_bit_size()
      replace_bit_size = replace.get_bit_size()
      cmp_result = self.compare_bitsizes(search_bit_size, replace_bit_size)

      assert cmp_result is not None and cmp_result <= 0, \
         'The search expression bit size {} and replace expression ' \
         'bit size {} may not be the same'.format(
               search_bit_size, replace_bit_size)

      replace.set_bit_size(search)

      self.validate_replace(replace, search)

_optimization_ids = itertools.count()

condition_list = ['true']

class SearchAndReplace(object):
   def __init__(self, transform, algebraic_pass):
      self.id = next(_optimization_ids)

      search = transform[0]
      replace = transform[1]
      if len(transform) > 2:
         self.condition = transform[2]
      else:
         self.condition = 'true'

      if self.condition not in condition_list:
         condition_list.append(self.condition)
      self.condition_index = condition_list.index(self.condition)

      varset = VarSet()
      if isinstance(search, Expression):
         self.search = search
      else:
         self.search = Expression(search, "search{0}".format(self.id), varset, algebraic_pass)

      varset.lock()

      if isinstance(replace, Value):
         self.replace = replace
      else:
         self.replace = Value.create(replace, "replace{0}".format(self.id), varset, algebraic_pass)

      BitSizeValidator(varset).validate(self.search, self.replace)

class TreeAutomaton(object):
   """This class calculates a bottom-up tree automaton to quickly search for
   the left-hand sides of tranforms. Tree automatons are a generalization of
   classical NFA's and DFA's, where the transition function determines the
   state of the parent node based on the state of its children. We construct a
   deterministic automaton to match patterns, using a similar algorithm to the
   classical NFA to DFA construction. At the moment, it only matches opcodes
   and constants (without checking the actual value), leaving more detailed
   checking to the search function which actually checks the leaves. The
   automaton acts as a quick filter for the search function, requiring only n
   + 1 table lookups for each n-source operation. The implementation is based
   on the theory described in "Tree Automatons: Two Taxonomies and a Toolkit."
   In the language of that reference, this is a frontier-to-root deterministic
   automaton using only symbol filtering. The filtering is crucial to reduce
   both the time taken to generate the tables and the size of the tables.
   """
   def __init__(self, transforms):
      self.patterns = [t.search for t in transforms]
      self._compute_items()
      self._build_table()
      #print('num items: {}'.format(len(set(self.items.values()))))
      #print('num states: {}'.format(len(self.states)))
      #for state, patterns in zip(self.states, self.patterns):
      #   print('{}: num patterns: {}'.format(state, len(patterns)))

   class IndexMap(object):
      """An indexed list of objects, where one can either lookup an object by
      index or find the index associated to an object quickly using a hash
      table. Compared to a list, it has a constant time index(). Compared to a
      set, it provides a stable iteration order.
      """
      def __init__(self, iterable=()):
         self.objects = []
         self.map = {}
         for obj in iterable:
            self.add(obj)

      def __getitem__(self, i):
         return self.objects[i]

      def __contains__(self, obj):
         return obj in self.map

      def __len__(self):
         return len(self.objects)

      def __iter__(self):
         return iter(self.objects)

      def clear(self):
         self.objects = []
         self.map.clear()

      def index(self, obj):
         return self.map[obj]

      def add(self, obj):
         if obj in self.map:
            return self.map[obj]
         else:
            index = len(self.objects)
            self.objects.append(obj)
            self.map[obj] = index
            return index

      def __repr__(self):
         return 'IndexMap([' + ', '.join(repr(e) for e in self.objects) + '])'

   class Item(object):
      """This represents an "item" in the language of "Tree Automatons." This
      is just a subtree of some pattern, which represents a potential partial
      match at runtime. We deduplicate them, so that identical subtrees of
      different patterns share the same object, and store some extra
      information needed for the main algorithm as well.
      """
      def __init__(self, opcode, children):
         self.opcode = opcode
         self.children = children
         # These are the indices of patterns for which this item is the root node.
         self.patterns = []
         # This the set of opcodes for parents of this item. Used to speed up
         # filtering.
         self.parent_ops = set()

      def __str__(self):
         return '(' + ', '.join([self.opcode] + [str(c) for c in self.children]) + ')'

      def __repr__(self):
         return str(self)

   def _compute_items(self):
      """Build a set of all possible items, deduplicating them."""
      # This is a map from (opcode, sources) to item.
      self.items = {}

      # The set of all opcodes used by the patterns. Used later to avoid
      # building and emitting all the tables for opcodes that aren't used.
      self.opcodes = self.IndexMap()

      def get_item(opcode, children, pattern=None):
         commutative = len(children) >= 2 \
               and "2src_commutative" in opcodes[opcode].algebraic_properties
         item = self.items.setdefault((opcode, children),
                                      self.Item(opcode, children))
         if commutative:
            self.items[opcode, (children[1], children[0]) + children[2:]] = item
         if pattern is not None:
            item.patterns.append(pattern)
         return item

      self.wildcard = get_item("__wildcard", ())
      self.const = get_item("__const", ())

      def process_subpattern(src, pattern=None):
         if isinstance(src, Constant):
            # Note: we throw away the actual constant value!
            return self.const
         elif isinstance(src, Variable):
            if src.is_constant:
               return self.const
            else:
               # Note: we throw away which variable it is here! This special
               # item is equivalent to nu in "Tree Automatons."
               return self.wildcard
         else:
            assert isinstance(src, Expression)
            opcode = src.opcode
            stripped = opcode.rstrip('0123456789')
            if stripped in conv_opcode_types:
               # Matches that use conversion opcodes with a specific type,
               # like f2i1, are tricky.  Either we construct the automaton to
               # match specific NIR opcodes like nir_op_f2i1, in which case we
               # need to create separate items for each possible NIR opcode
               # for patterns that have a generic opcode like f2i, or we
               # construct it to match the search opcode, in which case we
               # need to map f2i1 to f2i when constructing the automaton. Here
               # we do the latter.
               opcode = stripped
            self.opcodes.add(opcode)
            children = tuple(process_subpattern(c) for c in src.sources)
            item = get_item(opcode, children, pattern)
            for i, child in enumerate(children):
               child.parent_ops.add(opcode)
            return item

      for i, pattern in enumerate(self.patterns):
         process_subpattern(pattern, i)

   def _build_table(self):
      """This is the core algorithm which builds up the transition table. It
      is based off of Algorithm 5.7.38 "Reachability-based tabulation of Cl .
      Comp_a and Filt_{a,i} using integers to identify match sets." It
      simultaneously builds up a list of all possible "match sets" or
      "states", where each match set represents the set of Item's that match a
      given instruction, and builds up the transition table between states.
      """
      # Map from opcode + filtered state indices to transitioned state.
      self.table = defaultdict(dict)
      # Bijection from state to index. q in the original algorithm is
      # len(self.states)
      self.states = self.IndexMap()
      # Lists of pattern matches separated by None
      self.state_patterns = [None]
      # Offset in the ->transforms table for each state index
      self.state_pattern_offsets = []
      # Map from state index to filtered state index for each opcode.
      self.filter = defaultdict(list)
      # Bijections from filtered state to filtered state index for each
      # opcode, called the "representor sets" in the original algorithm.
      # q_{a,j} in the original algorithm is len(self.rep[op]).
      self.rep = defaultdict(self.IndexMap)

      # Everything in self.states with a index at least worklist_index is part
      # of the worklist of newly created states. There is also a worklist of
      # newly fitered states for each opcode, for which worklist_indices
      # serves a similar purpose. worklist_index corresponds to p in the
      # original algorithm, while worklist_indices is p_{a,j} (although since
      # we only filter by opcode/symbol, it's really just p_a).
      self.worklist_index = 0
      worklist_indices = defaultdict(lambda: 0)

      # This is the set of opcodes for which the filtered worklist is non-empty.
      # It's used to avoid scanning opcodes for which there is nothing to
      # process when building the transition table. It corresponds to new_a in
      # the original algorithm.
      new_opcodes = self.IndexMap()

      # Process states on the global worklist, filtering them for each opcode,
      # updating the filter tables, and updating the filtered worklists if any
      # new filtered states are found. Similar to ComputeRepresenterSets() in
      # the original algorithm, although that only processes a single state.
      def process_new_states():
         while self.worklist_index < len(self.states):
            state = self.states[self.worklist_index]
            # Calculate pattern matches for this state. Each pattern is
            # assigned to a unique item, so we don't have to worry about
            # deduplicating them here. However, we do have to sort them so
            # that they're visited at runtime in the order they're specified
            # in the source.
            patterns = list(sorted(p for item in state for p in item.patterns))

            if patterns:
                # Add our patterns to the global table.
                self.state_pattern_offsets.append(len(self.state_patterns))
                self.state_patterns.extend(patterns)
                self.state_patterns.append(None)
            else:
                # Point to the initial sentinel in the global table.
                self.state_pattern_offsets.append(0)

            # calculate filter table for this state, and update filtered
            # worklists.
            for op in self.opcodes:
               filt = self.filter[op]
               rep = self.rep[op]
               filtered = frozenset(item for item in state if \
                  op in item.parent_ops)
               if filtered in rep:
                  rep_index = rep.index(filtered)
               else:
                  rep_index = rep.add(filtered)
                  new_opcodes.add(op)
               assert len(filt) == self.worklist_index
               filt.append(rep_index)
            self.worklist_index += 1

      # There are two start states: one which can only match as a wildcard,
      # and one which can match as a wildcard or constant. These will be the
      # states of intrinsics/other instructions and load_const instructions,
      # respectively. The indices of these must match the definitions of
      # WILDCARD_STATE and CONST_STATE below, so that the runtime C code can
      # initialize things correctly.
      self.states.add(frozenset((self.wildcard,)))
      self.states.add(frozenset((self.const,self.wildcard)))
      process_new_states()

      while len(new_opcodes) > 0:
         for op in new_opcodes:
            rep = self.rep[op]
            table = self.table[op]
            op_worklist_index = worklist_indices[op]
            if op in conv_opcode_types:
               num_srcs = 1
            else:
               num_srcs = opcodes[op].num_inputs

            # Iterate over all possible source combinations where at least one
            # is on the worklist.
            for src_indices in itertools.product(range(len(rep)), repeat=num_srcs):
               if all(src_idx < op_worklist_index for src_idx in src_indices):
                  continue

               srcs = tuple(rep[src_idx] for src_idx in src_indices)

               # Try all possible pairings of source items and add the
               # corresponding parent items. This is Comp_a from the paper.
               parent = set(self.items[op, item_srcs] for item_srcs in
                  itertools.product(*srcs) if (op, item_srcs) in self.items)

               # We could always start matching something else with a
               # wildcard. This is Cl from the paper.
               parent.add(self.wildcard)

               table[src_indices] = self.states.add(frozenset(parent))
            worklist_indices[op] = len(rep)
         new_opcodes.clear()
         process_new_states()

_algebraic_pass_template = mako.template.Template("""
#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following ${len(xforms)}
 * transforms:
% for xform in xforms:
 *    ${xform.search} => ${xform.replace}
% endfor
 */

<% cache = {"next_index": 0} %>
static const nir_search_value_union ${pass_name}_values[] = {
% for xform in xforms:
   /* ${xform.search} => ${xform.replace} */
${xform.search.render(cache)}
${xform.replace.render(cache)}
% endfor
};

% if expression_cond:
static const nir_search_expression_cond ${pass_name}_expression_cond[] = {
% for cond in expression_cond:
   ${cond[0]},
% endfor
};
% endif

% if variable_cond:
static const nir_search_variable_cond ${pass_name}_variable_cond[] = {
% for cond in variable_cond:
   ${cond[0]},
% endfor
};
% endif

static const struct transform ${pass_name}_transforms[] = {
% for i in automaton.state_patterns:
% if i is not None:
   { ${xforms[i].search.array_index}, ${xforms[i].replace.array_index}, ${xforms[i].condition_index} },
% else:
   { ~0, ~0, ~0 }, /* Sentinel */

% endif
% endfor
};

static const struct per_op_table ${pass_name}_pass_op_table[nir_num_search_ops] = {
% for op in automaton.opcodes:
   [${get_c_opcode(op)}] = {
% if all(e == 0 for e in automaton.filter[op]):
      .filter = NULL,
% else:
      .filter = (const uint16_t []) {
      % for e in automaton.filter[op]:
         ${e},
      % endfor
      },
% endif
      <%
        num_filtered = len(automaton.rep[op])
      %>
      .num_filtered_states = ${num_filtered},
      .table = (const uint16_t []) {
      <%
        num_srcs = len(next(iter(automaton.table[op])))
      %>
      % for indices in itertools.product(range(num_filtered), repeat=num_srcs):
         ${automaton.table[op][indices]},
      % endfor
      },
   },
% endfor
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t ${pass_name}_transform_offsets[] = {
% for offset in automaton.state_pattern_offsets:
   ${offset},
% endfor
};

static const nir_algebraic_table ${pass_name}_table = {
   .transforms = ${pass_name}_transforms,
   .transform_offsets = ${pass_name}_transform_offsets,
   .pass_op_table = ${pass_name}_pass_op_table,
   .values = ${pass_name}_values,
   .expression_cond = ${ pass_name + "_expression_cond" if expression_cond else "NULL" },
   .variable_cond = ${ pass_name + "_variable_cond" if variable_cond else "NULL" },
};

bool
${pass_name}(
   nir_shader *shader
% for type, name in params:
   , ${type} ${name}
% endfor
) {
   bool progress = false;
   bool condition_flags[${len(condition_list)}];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(${str(cache["next_index"])} == ARRAY_SIZE(${pass_name}_values));
   % for index, condition in enumerate(condition_list):
   condition_flags[${index}] = ${condition};
   % endfor

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &${pass_name}_table);
   }

   return progress;
}
""")


class AlgebraicPass(object):
   # params is a list of `("type", "name")` tuples
   def __init__(self, pass_name, transforms, params=[]):
      self.xforms = []
      self.opcode_xforms = defaultdict(lambda : [])
      self.pass_name = pass_name
      self.expression_cond = {}
      self.variable_cond = {}
      self.params = params

      error = False

      for xform in transforms:
         if not isinstance(xform, SearchAndReplace):
            try:
               xform = SearchAndReplace(xform, self)
            except:
               print("Failed to parse transformation:", file=sys.stderr)
               print("  " + str(xform), file=sys.stderr)
               traceback.print_exc(file=sys.stderr)
               print('', file=sys.stderr)
               error = True
               continue

         self.xforms.append(xform)
         if xform.search.opcode in conv_opcode_types:
            dst_type = conv_opcode_types[xform.search.opcode]
            for size in type_sizes(dst_type):
               sized_opcode = xform.search.opcode + str(size)
               self.opcode_xforms[sized_opcode].append(xform)
         else:
            self.opcode_xforms[xform.search.opcode].append(xform)

         # Check to make sure the search pattern does not unexpectedly contain
         # more commutative expressions than match_expression (nir_search.c)
         # can handle.
         comm_exprs = xform.search.comm_exprs

         if xform.search.many_commutative_expressions:
            if comm_exprs <= nir_search_max_comm_ops:
               print("Transform expected to have too many commutative " \
                     "expression but did not " \
                     "({} <= {}).".format(comm_exprs, nir_search_max_comm_op),
                     file=sys.stderr)
               print("  " + str(xform), file=sys.stderr)
               traceback.print_exc(file=sys.stderr)
               print('', file=sys.stderr)
               error = True
         else:
            if comm_exprs > nir_search_max_comm_ops:
               print("Transformation with too many commutative expressions " \
                     "({} > {}).  Modify pattern or annotate with " \
                     "\"many-comm-expr\".".format(comm_exprs,
                                                  nir_search_max_comm_ops),
                     file=sys.stderr)
               print("  " + str(xform.search), file=sys.stderr)
               print("{}".format(xform.search.cond), file=sys.stderr)
               error = True

      self.automaton = TreeAutomaton(self.xforms)

      if error:
         sys.exit(1)


   def render(self):
      return _algebraic_pass_template.render(pass_name=self.pass_name,
                                             xforms=self.xforms,
                                             opcode_xforms=self.opcode_xforms,
                                             condition_list=condition_list,
                                             automaton=self.automaton,
                                             expression_cond = sorted(self.expression_cond.items(), key=lambda kv: kv[1]),
                                             variable_cond = sorted(self.variable_cond.items(), key=lambda kv: kv[1]),
                                             get_c_opcode=get_c_opcode,
                                             itertools=itertools,
                                             params=self.params)

# The replacement expression isn't necessarily exact if the search expression is exact.
def ignore_exact(*expr):
   expr = SearchExpression.create(expr)
   expr.ignore_exact = True
   return expr
